#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
import os
import pickle
import sys
import time
import unittest
from multiprocessing.pool import ThreadPool
from numpy import mean, array, concatenate, random, testing
from psutil import cpu_count

from dlg import droputils, drop_loaders
from dlg.apps.simple import (
    GenericScatterApp,
    GenericNpyScatterApp,
    ListAppendThrashingApp,
)
from dlg.apps.simple import RandomArrayApp, AverageArraysApp, HelloWorldApp
from dlg.apps.simple import SleepApp, CopyApp, SleepAndCopyApp
from dlg.ddap_protocol import DROPStates
from dlg.data.drops.data_base import NullDROP
from dlg.data.drops.memory import InMemoryDROP
from dlg.data.drops.ngas import NgasDROP
from dlg.data.drops.file import FileDROP

if sys.version_info >= (3, 8):
    from dlg.manager.shared_memory_manager import DlgSharedMemoryManager

logger = logging.getLogger(__name__)


class TestSimpleApps(unittest.TestCase):
    def _test_graph_runs(self, drops, first, last, timeout=300):
        first = droputils.listify(first)
        with droputils.DROPWaiterCtx(self, last, timeout):
            for f in first:
                f.setCompleted()

        for x in drops:
            self.assertEqual(DROPStates.COMPLETED, x.status)

    def test_sleepapp(self):
        # Nothing fancy, just run it and be done with it
        a = NullDROP("a", "a")
        b = SleepApp("b", "b")
        c = NullDROP("c", "c")
        b.addInput(a)
        b.addOutput(c)

        self._test_graph_runs((a, b, c), a, c)

    def _test_copyapp_simple(self, app):
        # Again, not foo fancy, simple apps require simple tests
        a, c = (InMemoryDROP(x, x, nm=x) for x in ("a", "c"))
        b = app("b", "b")
        b.addInput(a)
        b.addOutput(c)

        data = os.urandom(32)
        a.write(data)

        self._test_graph_runs((a, b, c), a, c)
        self.assertEqual(data, droputils.allDropContents(c))

    def _test_copyapp_order_preserved(self, app):
        # Inputs are copied in the order they are added
        a, b, d = (InMemoryDROP(x, x, nm=x) for x in ("a", "b", "d"))
        c = app("c", "c")
        for x in a, b:
            c.addInput(x)
        c.addOutput(d)

        data1 = os.urandom(32)
        data2 = os.urandom(32)
        a.write(data1)
        b.write(data2)

        self._test_graph_runs((a, b, c, d), (a, b), d)
        self.assertEqual(data1 + data2, droputils.allDropContents(d))

    def _test_copyapp(self, app):
        self._test_copyapp_simple(app)
        self._test_copyapp_order_preserved(app)

    def test_copyapp(self):
        self._test_copyapp(CopyApp)

    def test_sleepandcopyapp(self):
        self._test_copyapp(SleepAndCopyApp)

    def test_randomarrayapp(self):
        i = NullDROP("i", "i")
        c = RandomArrayApp("c", "c", keep_array=True)
        o = InMemoryDROP("o", "o")
        c.addInput(i)
        c.addOutput(o)
        self._test_graph_runs((i, c, o), i, o)
        marray = c._getArray()
        data = pickle.loads(droputils.allDropContents(o))
        v = marray == data
        self.assertEqual(v.all(), True)

    def test_averagearraysapp(self):
        a = AverageArraysApp("a", "a")
        i1, i2, o = (InMemoryDROP(x, x) for x in ("i1", "i2", "o"))
        c = AverageArraysApp("c", "c")
        c.addInput(i1)
        c.addInput(i2)
        c.addOutput(o)
        for x in i1, i2:
            a.addInput(x)
        data1 = random.randint(0, 100, size=100)
        data2 = random.randint(0, 100, size=100)
        i1.write(pickle.dumps(data1))
        i2.write(pickle.dumps(data2))
        big_mean = mean(mean([array(data1), array(data2)], axis=0))
        self._test_graph_runs((i1, i2, c, o), (i1, i2), o)
        average = pickle.loads(droputils.allDropContents(o))
        self.assertEqual(big_mean, average)

    def test_helloworldapp(self):
        h = HelloWorldApp("h", "h")
        b = FileDROP("c", "c")
        h.addOutput(b)
        b.addProducer(h)
        h.execute()
        self.assertEqual(h.greeting.encode("utf8"), droputils.allDropContents(b))

    def test_parallelHelloWorld(self):
        m0 = InMemoryDROP("m0", "m0")
        s = GenericScatterApp("s", "s", num_of_copies=4)
        greets = ["World", "Solar system", "Galaxy", "Universe"]
        m0.write(pickle.dumps(greets))
        s.addInput(m0)
        m = []
        h = []
        f = []
        for i in range(1, len(greets) + 1, 1):
            m.append(InMemoryDROP("m%d" % i, "m%d" % i))
            h.append(HelloWorldApp("h%d" % i, "h%d" % i))
            f.append(FileDROP("f%d" % i, "f%d" % i))
            s.addOutput(m[-1])
            h[-1].addInput(m[-1])
            h[-1].addOutput(f[-1])
        ad = [m0, s]
        ad.extend(m)
        ad.extend(h)
        ad.extend(f)
        self._test_graph_runs(ad, m0, f)
        for i in range(len(f)):
            self.assertEqual(
                ("Hello %s" % greets[i]).encode("utf8"),
                droputils.allDropContents(f[i]),
            )

    @unittest.skip
    def test_ngasio(self):
        nd_in = NgasDROP("HelloWorld_out.txt", "HelloWorld_out.txt")
        nd_in.ngasSrv = "ngas.icrar.org"
        b = CopyApp("b", "b")
        did = "HelloWorld-%f" % time.time()
        nd_out = NgasDROP(did, did, len=11)
        nd_out.ngasSrv = "ngas.icrar.org"
        nd_out.len = nd_in.size
        d = CopyApp("d", "d")
        i = InMemoryDROP("i", "i")
        # b.addInput(nd_in)
        # b.addOutput(nd_out)
        nd_in.addConsumer(b)
        nd_out.addProducer(b)
        d.addInput(nd_out)
        i.addProducer(d)
        # b.addOutput(i)
        self._test_graph_runs((nd_in, b, nd_out, i, d), nd_in, i, timeout=60)
        self.assertEqual(b"Hello World", droputils.allDropContents(i))

    def test_genericScatter(self):
        data_in = random.randint(0, 100, size=100)
        b = InMemoryDROP("b", "b")
        b.write(pickle.dumps(data_in))
        s = GenericScatterApp("s", "s", num_of_copies=2)
        s.addInput(b)
        o1 = InMemoryDROP("o1", "o1")
        o2 = InMemoryDROP("o2", "o2")
        for x in o1, o2:
            s.addOutput(x)
        self._test_graph_runs((b, s, o1, o2), b, (o1, o2), timeout=4)

        data1 = pickle.loads(droputils.allDropContents(o1))
        data2 = pickle.loads(droputils.allDropContents(o2))
        data_out = concatenate([data1, data2])
        self.assertEqual(data_in.all(), data_out.all())

    def test_genericNpyScatter(self):
        data_in = random.rand(100, 100)
        b = InMemoryDROP("b", "b")
        drop_loaders.save_numpy(b, data_in)
        s = GenericNpyScatterApp("s", "s", num_of_copies=2)
        s.addInput(b)
        o1 = InMemoryDROP("o1", "o1")
        o2 = InMemoryDROP("o2", "o2")
        for x in o1, o2:
            s.addOutput(x)
        self._test_graph_runs((b, s, o1, o2), b, (o1, o2), timeout=4)

        data1 = drop_loaders.load_numpy(o1)
        data2 = drop_loaders.load_numpy(o2)
        data_out = concatenate([data1, data2])
        self.assertEqual(data_in.all(), data_out.all())

    def test_genericNpyScatter_multi(self):
        data1_in = random.rand(100, 100)
        data2_in = random.rand(100, 100)
        b = InMemoryDROP("b", "b")
        c = InMemoryDROP("c", "c")
        drop_loaders.save_numpy(b, data1_in)
        drop_loaders.save_numpy(c, data2_in)
        s = GenericNpyScatterApp("s", "s", num_of_copies=2, scatter_axes="[0,0]")
        s.addInput(b)
        s.addInput(c)
        o1 = InMemoryDROP("o1", "o1")
        o2 = InMemoryDROP("o2", "o2")
        o3 = InMemoryDROP("o3", "o3")
        o4 = InMemoryDROP("o4", "o4")
        for x in o1, o2, o3, o4:
            s.addOutput(x)
        self._test_graph_runs(
            (b, s, o1, o2, o3, o4), (b, c), (o1, o2, o3, o4), timeout=4
        )

        data11 = drop_loaders.load_numpy(o1)
        data12 = drop_loaders.load_numpy(o2)
        data1_out = concatenate([data11, data12])
        self.assertEqual(data1_out.shape, data1_in.shape)
        testing.assert_array_equal(data1_out, data1_in)

        data21 = drop_loaders.load_numpy(o3)
        data22 = drop_loaders.load_numpy(o4)
        data2_out = concatenate([data21, data22])
        testing.assert_array_equal(data2_out, data2_in)

    def test_listappendthrashing(self, size=1000):
        a = InMemoryDROP("a", "a")
        b = ListAppendThrashingApp("b", "b", size=size)
        self.assertEqual(b.size, size)
        c = InMemoryDROP("c", "c")
        b.addInput(a)
        b.addOutput(c)
        self._test_graph_runs((a, b, c), a, c, timeout=4)
        data_out = pickle.loads(droputils.allDropContents(c))
        self.assertEqual(b.marray, data_out)

    @unittest.skipIf(
        sys.version_info < (3, 8),
        "Multiprocessing not compatible with Python < 3.8",
    )
    def test_multi_listappendthrashing(self, size=1000, parallel=True):
        max_threads = cpu_count(logical=False)
        drop_ids = [chr(97 + x) for x in range(max_threads)]
        threadpool = ThreadPool(processes=max_threads)
        memory_manager = DlgSharedMemoryManager()
        session_id = 1
        memory_manager.register_session(session_id)
        S = InMemoryDROP("S", "S")
        X = AverageArraysApp("X", "X")
        Z = InMemoryDROP("Z", "Z")
        drops = [ListAppendThrashingApp(x, x, size=size) for x in drop_ids]
        mdrops = [InMemoryDROP(chr(65 + x), chr(65 + x)) for x in range(max_threads)]
        if parallel:
            # a bit of magic to get the app drops using the processes
            _ = [drop.__setattr__("_tp", threadpool) for drop in drops]
            _ = [drop.__setattr__("_tp", threadpool) for drop in mdrops]
            _ = [drop.__setattr__("_sessionId", session_id) for drop in mdrops]
            _ = [memory_manager.register_drop(drop.uid, session_id) for drop in mdrops]
            X.__setattr__("_tp", threadpool)
            Z.__setattr__("_tp", threadpool)
            Z.__setattr__("_sessionId", session_id)
            memory_manager.register_drop(Z.uid, session_id)

        _ = [d.addInput(S) for d in drops]
        _ = [d.addOutput(m) for d, m in zip(drops, mdrops)]
        _ = [X.addInput(m) for m in mdrops]
        X.addOutput(Z)
        logger.info(f"Number of inputs/outputs: {len(X.inputs)}, {len(X.outputs)}")
        self._test_graph_runs([S, X, Z] + drops + mdrops, S, Z, timeout=200)
        # Need to run our 'copy' of the averaging APP
        num_array = []
        for drop in mdrops:
            buf = droputils.allDropContents(drop)
            num_array.extend(pickle.loads(buf))
        X.marray = num_array
        average = X.averageArray()
        # Load actual results
        graph_result = droputils.allDropContents(Z)
        graph_result = pickle.loads(graph_result)
        self.assertEqual(graph_result, average)
        # Must be called to unlink all shared memory
        memory_manager.shutdown_all()

    # @unittest.skipIf(
    #     sys.version_info < (3, 8), "Multiprocessing not compatible with Python < 3.8"
    # )
    @unittest.skip
    def test_speedup(self):
        """
        Run serial and parallel test and report speedup.
        NOTE: In order to get the stdout you need to run pyest with
        --capture=tee-sys
        """
        size = 3000
        logger.info("Starting serial test..")
        st = time.time()
        self.test_multi_listappendthrashing(size=size, parallel=False)
        t1 = time.time() - st
        time.sleep(5)
        logger.info("Starting parallel test..")
        st = time.time()
        self.test_multi_listappendthrashing(size=size, parallel=True)
        t2 = time.time() - st
        logger.info(f"Speedup: {t1 / t2:.2f} from {cpu_count(logical=False)} cores")
        # TODO: This is unpredictable, but maybe we can do something meaningful.
        # self.assertAlmostEqual(t1/cpu_count(logical=False), t2, 1)
        # How about this? We only need to see some type of speedup
        if cpu_count(logical=False) > 1 and size > 500:
            self.assertGreater(t1 / t2, 1)
        else:
            # Ensure that multi-threading overhead doesn't ruin serial performance?
            self.assertAlmostEqual(t1, t2, delta=0.5)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Tests the low-level functionality for drops to hash runtime data.
"""

import unittest

from dlg.common.reproducibility.constants import ReproducibilityFlags
from dlg.common.reproducibility.reproducibility import common_hash
from dlg.ddap_protocol import DROPStates
from dlg.drop import AbstractDROP
from merklelib import MerkleTree


class AbstractDROPHashTests(unittest.TestCase):
    """
    Tests the reprodata generation methods of the AbstractDROP class
    """

    def test_null_merkleroot(self):
        """
        Sanity check that the default MerkleRoot of an abstract drop is Null
        Consider it a cardinal sin to change this.
        """
        drop_a = AbstractDROP("a", "a")
        self.assertIsNone(drop_a.merkleroot)

    def test_generate_rerun_data(self):
        """
        Tests that completed Rerun data contains the completed flag.
        """
        drop_a = AbstractDROP("a", "a")
        drop_a.reproducibility_level = ReproducibilityFlags.RERUN
        drop_a.setCompleted()
        self.assertEqual(drop_a.generate_rerun_data(), {"status": DROPStates.COMPLETED})

    def test_commit_on_complete(self):
        """
        Tests that merkle_data is generated upon set_complete status and is correct (NOTHING, RERUN)
        """

        drop_a = AbstractDROP("a", "a")
        drop_b = AbstractDROP("b", "b")
        drop_a.reproducibility_level = ReproducibilityFlags.RERUN
        drop_b.reproducibility_level = ReproducibilityFlags.NOTHING
        self.assertIsNone(drop_a.merkleroot)
        self.assertIsNone(drop_b.merkleroot)

        # Test RERUN
        drop_a.setCompleted()
        test = MerkleTree({"status": DROPStates.COMPLETED}.items(), common_hash)
        # 689fcf0d74c42200bef177db545adc43c135dfb0d7dc85b166db3af1dcded235
        self.assertTrue(test.merkle_root == drop_a.merkleroot)

        # Test NOTHING
        drop_b.setCompleted()
        # None
        self.assertIsNone(drop_b.merkleroot)
        self.assertTrue(drop_b._committed)

    def test_set_reproducibility_level(self):
        """
        Tests functionality for changing a DROP's reproducibility flag
        If already committed. The drop should reset and re-commit all reproducibility data
        If not committed, the change can proceed simply.
        """
        drop_a = AbstractDROP("a", "a")
        drop_b = AbstractDROP("b", "b")
        drop_a.reproducibility_level = ReproducibilityFlags.NOTHING
        drop_b.reproducibility_level = ReproducibilityFlags.NOTHING

        drop_a.setCompleted()
        self.assertIsNone(drop_a.merkleroot)
        drop_a.reproducibility_level = ReproducibilityFlags.RERUN
        drop_a.commit()
        self.assertIsNotNone(drop_a.merkleroot)

        self.assertIsNone(drop_b.merkleroot)
        drop_b.reproducibility_level = ReproducibilityFlags.RERUN
        drop_b.setCompleted()
        self.assertIsNotNone(drop_b.merkleroot)

        with self.assertRaises(TypeError):
            drop_a.reproducibility_level = "REPEAT"

    def test_set_all(self):
        """
        Tests setting the reproducibility setting to ALL
        """
        drop_a = AbstractDROP("a", "a")
        self.assertIsNone(drop_a.merkleroot)
        drop_a.reproducibility_level = ReproducibilityFlags.ALL
        self.assertTrue(isinstance(drop_a.merkleroot, dict))

    def test_set_all_set_rerun(self):
        """
        Tests taking a drop that was set to ALL to be set to RERUN and have data structures
        reset correctly.
        """
        drop_a = AbstractDROP("a", "a")
        self.assertIsNone(drop_a.merkleroot)
        drop_a.reproducibility_level = ReproducibilityFlags.ALL
        drop_a.commit()
        self.assertTrue(isinstance(drop_a.merkleroot, dict))
        drop_a.reproducibility_level = ReproducibilityFlags.RERUN
        drop_a.commit()
        self.assertTrue(isinstance(drop_a.merkleroot, str))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import copy
import os
import sys
import threading
import unittest
from time import sleep
import multiprocessing
import random

from dlg import droputils
from dlg.common import dropdict
from dlg.ddap_protocol import DROPStates, DROPRel, DROPLinkType
from dlg.apps.app_base import BarrierAppDROP
from dlg.manager.node_manager import NodeManager

try:
    from crc32c import crc32c  # @UnusedImport
except:
    from binascii import crc32  # @Reimport

random.seed(42)

hostname = "localhost"
default_repro = {
    "rmode": "1",
    "RERUN": {
        "lg_blockhash": "x",
        "pgt_blockhash": "y",
        "pg_blockhash": "z",
    },
}
default_graph_repro = {
    "rmode": "1",
    "meta_data": {"repro_protocol": 0.1, "hashing_alg": "_sha3.sha3_256"},
    "merkleroot": "a",
    "RERUN": {
        "signature": "b",
    },
}


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = default_repro.copy()
    graph.append(default_graph_repro.copy())
    return graph


def memory(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": uid,
            "categoryType": "Data",
            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            "reprodata": default_repro.copy(),
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def sleepAndCopy(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": uid,
            "categoryType": "Application",
            "dropclass": "dlg.apps.simple.SleepAndCopyApp",
            "reprodata": default_repro.copy(),
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def quickDeploy(nm, sessionId, graphSpec, node_subscriptions={}):
    nm.createSession(sessionId)
    nm.addGraphSpec(sessionId, graphSpec)
    nm.add_node_subscriptions(sessionId, node_subscriptions)
    nm.deploySession(sessionId)


class ErroneousApp(BarrierAppDROP):
    def run(self):
        raise Exception("Sorry, we always fail")


def nm_conninfo(n):
    return "localhost", 5553 + n, 6666 + n


class NMTestsMixIn(object):
    def __init__(self, *args, **kwargs):
        super(NMTestsMixIn, self).__init__(*args, **kwargs)
        self._dms = []
        self.use_processes = False

    def _start_dm(self, threads=0, **kwargs):
        host, events_port, rpc_port = nm_conninfo(len(self._dms))
        nm = NodeManager(
            host=host,
            events_port=events_port,
            rpc_port=rpc_port,
            max_threads=threads,
            use_processes=self.use_processes,
            **kwargs,
        )
        self._dms.append(nm)
        return nm

    def tearDown(self):
        super(NMTestsMixIn, self).tearDown()
        for nm in self._dms:
            nm.shutdown()

    def _test_runGraphInTwoNMs(
        self,
        g1,
        g2,
        rels,
        root_data,
        leaf_data,
        root_oids=("A",),
        leaf_oid="C",
        expected_failures=[],
        sessionId=f"s{random.randint(0, 1000)}",
        node_managers=None,
        threads=0,
    ):
        """Utility to run a graph in two Node Managers"""

        dm1, dm2 = node_managers or [
            self._start_dm(threads=threads) for _ in range(2)
        ]
        add_test_reprodata(g1)
        add_test_reprodata(g2)
        quickDeploy(dm1, sessionId, g1, {nm_conninfo(1): rels})
        quickDeploy(dm2, sessionId, g2, {nm_conninfo(0): rels})
        self.assertEqual(len(g1), len(dm1._sessions[sessionId].drops))
        self.assertEqual(len(g2), len(dm2._sessions[sessionId].drops))

        # Run! We wait until c is completed
        drops = {}
        drops.update(dm1._sessions[sessionId].drops)
        drops.update(dm2._sessions[sessionId].drops)

        leaf_drop = drops[leaf_oid]
        with droputils.DROPWaiterCtx(self, leaf_drop, 5):
            for oid in root_oids:
                drop = drops[oid]
                drop.write(root_data)
                drop.setCompleted()

        expected_successes = [
            drops[oid] for oid in drops if oid not in expected_failures
        ]
        expected_failures = [
            drops[oid] for oid in drops if oid in expected_failures
        ]
        for drop in expected_successes:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        for drop in expected_failures:
            self.assertEqual(DROPStates.ERROR, drop.status)

        leaf_drop_data = None
        if leaf_drop not in expected_failures:
            leaf_drop_data = droputils.allDropContents(leaf_drop)
            if leaf_data is not None:
                self.assertEqual(len(leaf_data), len(leaf_drop_data))
                self.assertEqual(leaf_data, leaf_drop_data)

        sleep(0.1)  # just make sure all events have been processed.
        dm1.destroySession(sessionId)
        dm2.destroySession(sessionId)
        return leaf_drop_data


class NodeManagerTestsBase(NMTestsMixIn):
    def _deploy_error_graph(self, **kwargs):
        sessionId = f"s{random.randint(0, 1000)}"
        g = [
            memory("A"),
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "test.manager.test_dm.ErroneousApp",
                "inputs": ["A"],
            },
            memory("C", producers=["B"]),
        ]
        add_test_reprodata(g)
        dm = self._start_dm(**kwargs)
        dm.createSession(sessionId)
        dm.addGraphSpec(sessionId, g)
        dm.deploySession(sessionId, ["A"])

    def test_error_listener(self):
        evt = threading.Event()
        erroneous_drops = []

        class listener(object):
            def on_error(self, drop):
                erroneous_drops.append(drop.uid)
                if (
                    len(erroneous_drops) == 2
                ):  # both 'C' and 'B' failed already
                    evt.set()

        self._deploy_error_graph(error_listener=listener())
        self.assertTrue(evt.wait(10), "Didn't receive errors on time")

    def test_event_listener(self):
        """Tests that user-provided event listeners work"""

        evt = threading.Event()

        class listener(object):
            def __init__(self):
                self.recv = 0

            def handleEvent(self, _evt):
                self.recv += 1
                if self.recv == 3:
                    evt.set()

        self._deploy_error_graph(event_listeners=[listener()])
        self.assertTrue(evt.wait(10), "Didn't receive events on time")

    def _test_runGraphOneDOPerDOM(self, repeats=1):
        g1 = [memory("A")]
        g2 = [
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.crc.CRCApp",
            },
            memory("C", producers=["B"]),
        ]
        rels = [DROPRel("B", DROPLinkType.CONSUMER, "A")]
        a_data = os.urandom(32)
        c_data = str(crc32c(a_data, 0)).encode("utf8")
        node_managers = [
            self._start_dm() for _ in range(2)
        ]
        ids = [0] * repeats
        for n in range(repeats):
            choice = 0
            while choice in ids:
                choice = random.randint(0, 1000)
            ids[n] = choice
            sessionId = f"s{choice}"
            self._test_runGraphInTwoNMs(
                copy.deepcopy(g1),
                copy.deepcopy(g2),
                rels,
                a_data,
                c_data,
                sessionId=sessionId,
                node_managers=node_managers,
            )

    def test_runGraphOneDOPerDOM(self):
        """
        A test that creates three DROPs in two different DMs and runs the graph.
        For this the graphs that are fed into the DMs must *not* express the
        inter-DM relationships, although they are still passed down
        separately. The graph looks like:

        DM #1      DM #2
        =======    =============
        | A --|----|-> B --> C |
        =======    =============
        """
        self._test_runGraphOneDOPerDOM()

    def test_runGraphOneDOPerDOMTwice(self):
        """Like test_runGraphOneDOPerDOM but runs two sessions succesively"""
        self._test_runGraphOneDOPerDOM(2)

    def test_runGraphSeveralDropsPerDM(self):
        """
        A test that creates several DROPs in two different DMs and  runs
        the graph. The graph looks like this

        DM #1                  DM #2
        ===================    ================
        | A --> C --> D --|----|-|            |
        |                 |    | |--> E --> F |
        | B --------------|----|-|            |
        ===================    ================

        :see: `self.test_runGraphSingleDOPerDOM`
        """
        dm1, dm2 = [self._start_dm() for _ in range(2)]

        sessionId = "s1"
        g1 = [
            memory("A", consumers=["C"]),
            memory("B"),
            {
                "oid": "C",
                "categoryType": "Application",
                "dropclass": "dlg.apps.crc.CRCApp",
            },
            memory("D", producers=["C"]),
        ]
        g2 = [
            {
                "oid": "E",
                "categoryType": "Application",
                "dropclass": "test.test_drop.SumupContainerChecksum",
            },
            memory("F", producers=["E"]),
        ]
        add_test_reprodata(g1)
        add_test_reprodata(g2)
        rels = [
            DROPRel("D", DROPLinkType.INPUT, "E"),
            DROPRel("B", DROPLinkType.INPUT, "E"),
        ]
        quickDeploy(dm1, sessionId, g1, {nm_conninfo(1): rels})
        quickDeploy(dm2, sessionId, g2, {nm_conninfo(0): rels})

        self.assertEqual(4, len(dm1._sessions[sessionId].drops))
        self.assertEqual(2, len(dm2._sessions[sessionId].drops))

        # Run! The sole fact that this doesn't throw exceptions is already
        # a good proof that everything is working as expected
        a, b, c, d = [
            dm1._sessions[sessionId].drops[x] for x in ("A", "B", "C", "D")
        ]
        e, f = [dm2._sessions[sessionId].drops[x] for x in ("E", "F")]
        with droputils.DROPWaiterCtx(self, f, 5):
            a.write(b"a")
            a.setCompleted()
            b.write(b"a")
            b.setCompleted()

        for drop in a, b, c, d, e, f:
            self.assertEqual(
                DROPStates.COMPLETED,
                drop.status,
                "DROP %s is not COMPLETED" % (drop.uid),
            )

        self.assertEqual(a.checksum, int(droputils.allDropContents(d)))
        self.assertEqual(
            b.checksum + d.checksum, int(droputils.allDropContents(f))
        )

        dm1.destroySession(sessionId)
        dm2.destroySession(sessionId)

    def test_runWithFourDMs(self):
        """
        A test that creates several DROPs in two different DMs and  runs
        the graph. The graph looks like this

                      DM #2
                     +--------------------------+
                     |        |--> C --|        |
                 +---|--> B --|--> D --|--> F --|--|
                 |   |        |--> E --|        |  |
        DM #1    |   +--------------------------+  |   DM #4
        +-----+  |                                 |  +---------------------+
        |     |  |                                 |--|--> L --|            |
        | A --|--+                                    |        |--> N --> O |
        |     |  |                                 |--|--> M --|            |
        +-----+  |    DM #3                        |  +---------------------+
                 |   +--------------------------+  |
                 |   |        |--> H --|        |  |
                 +---|--> G --|--> I --|--> K --|--|
                     |        |--> J --|        |
                     +--------------------------+

        B, F, G, K and N are AppDOs; the rest are plain in-memory DROPs
        """

        dm1, dm2, dm3, dm4 = [
            self._start_dm() for _ in range(4)
        ]

        sessionId = f"s{random.randint(0, 1000)}"
        g1 = [memory("A", expectedSize=1)]
        g2 = [
            sleepAndCopy("B", outputs=["C", "D", "E"], sleepTime=0),
            memory("C"),
            memory("D"),
            memory("E"),
            sleepAndCopy("F", inputs=["C", "D", "E"], sleepTime=0),
        ]
        g3 = [
            sleepAndCopy("G", outputs=["H", "I", "J"], sleepTime=0),
            memory("H"),
            memory("I"),
            memory("J"),
            sleepAndCopy("K", inputs=["H", "I", "J"], sleepTime=0),
        ]
        g4 = [
            memory("L"),
            memory("M"),
            sleepAndCopy("N", inputs=["L", "M"], outputs=["O"], sleepTime=0),
            memory("O"),
        ]
        for g in [g1, g2, g3, g4]:
            add_test_reprodata(g)
        rels_12 = [DROPRel("A", DROPLinkType.INPUT, "B")]
        rels_13 = [DROPRel("A", DROPLinkType.INPUT, "G")]
        rels_24 = [DROPRel("F", DROPLinkType.PRODUCER, "L")]
        rels_34 = [DROPRel("K", DROPLinkType.PRODUCER, "M")]
        quickDeploy(
            dm1,
            sessionId,
            g1,
            {nm_conninfo(1): rels_12, nm_conninfo(2): rels_13},
        )
        quickDeploy(
            dm2,
            sessionId,
            g2,
            {nm_conninfo(0): rels_12, nm_conninfo(3): rels_24},
        )
        quickDeploy(
            dm3,
            sessionId,
            g3,
            {nm_conninfo(0): rels_13, nm_conninfo(3): rels_34},
        )
        quickDeploy(
            dm4,
            sessionId,
            g4,
            {nm_conninfo(1): rels_24, nm_conninfo(2): rels_34},
        )

        self.assertEqual(1, len(dm1._sessions[sessionId].drops))
        self.assertEqual(5, len(dm2._sessions[sessionId].drops))
        self.assertEqual(5, len(dm3._sessions[sessionId].drops))
        self.assertEqual(4, len(dm4._sessions[sessionId].drops))

        a = dm1._sessions[sessionId].drops["A"]
        o = dm4._sessions[sessionId].drops["O"]
        drops = []
        for x in (dm1, dm2, dm3, dm4):
            drops += x._sessions[sessionId].drops.values()

        # Run! This should trigger the full execution of the graph
        with droputils.DROPWaiterCtx(self, o, 5):
            a.write(b"a")

        for drop in drops:
            self.assertEqual(
                DROPStates.COMPLETED,
                drop.status,
                "Status of '%s' is not COMPLETED: %d"
                % (drop.uid, drop.status),
            )

        for dm in [dm1, dm2, dm3, dm4]:
            dm.destroySession(sessionId)

    def test_many_relationships(self):
        """
        A test in which a drop is related to many other drops that live in a
        separate DM.

        Drop A is accessed by many applications (B1, B2, .., BN), which should
        not exhaust resources on DM #1. We collapse all into C so we can monitor
        only its status to know that the execution is over.

        DM #1                     DM #2
        =======    ====================
        |     |    | |--> B1 --|      |
        |     |    | |--> B2 --|      |
        | A --|----|-|--> B3 --|--> C |
        |     |    | |.........|      |
        |     |    | |--> BN --|      |
        =======    ====================
        """

        dm1, dm2 = [self._start_dm() for _ in range(2)]

        sessionId = f"s{random.randint(0, 1000)}"
        N = 100
        g1 = [memory("A")]
        g2 = [memory("C")]
        rels = []
        for i in range(N):
            b_oid = "B%d" % (i,)
            # SleepAndCopyApp effectively opens the input drop
            g2.append(sleepAndCopy(b_oid, outputs=["C"], sleepTime=0))
            rels.append(DROPRel("A", DROPLinkType.INPUT, b_oid))
        add_test_reprodata(g1)
        add_test_reprodata(g2)
        quickDeploy(dm1, sessionId, g1, {nm_conninfo(1): rels})
        quickDeploy(dm2, sessionId, g2, {nm_conninfo(0): rels})
        self.assertEqual(1, len(dm1._sessions[sessionId].drops))
        self.assertEqual(1 + N, len(dm2._sessions[sessionId].drops))

        # Run! The sole fact that this doesn't throw exceptions is already
        # a good proof that everything is working as expected
        a = dm1._sessions[sessionId].drops["A"]
        c = dm2._sessions[sessionId].drops["C"]
        with droputils.DROPWaiterCtx(self, c, 10):
            a.write(b"a")
            a.setCompleted()

        for i in range(N):
            drop = dm2._sessions[sessionId].drops["B%d" % (i,)]
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        dm1.destroySession(sessionId)
        dm2.destroySession(sessionId)

    def test_runGraphSeveralDropsPerDM_with_get_consumer_nodes(self):
        """
        A test that creates several DROPs in two different DMs and runs
        the graph. Checks the node address(s) of the consumers in the second DM.
        The graph looks like this

        DM #1                  DM #2
        ===================    ================
        | A --> C --> D --|----|-| --> E      |
        |                 |    | |
        |                 |    | | --> F      |
        ===================    ================

        :see: `self.test_runGraphSeveralDropsPerDM_with_get_consumer_nodes`
        """
        ip_addr_1 = "8.8.8.8"
        ip_addr_2 = "8.8.8.9"

        dm1, dm2 = [self._start_dm() for _ in range(2)]

        sessionId = f"s{random.randint(0, 1000)}"
        g1 = [
            memory("A", consumers=["C"]),
            {
                "oid": "C",
                "categoryType": "Application",
                "dropclass": "dlg.apps.crc.CRCApp",
                "consumers": ["D"],
            },
            memory("D", producers=["C"]),
        ]
        g2 = [
            {
                "oid": "E",
                "categoryType": "Application",
                "dropclass": "test.test_drop.SumupContainerChecksum",
                "node": ip_addr_1,
            },
            {
                "oid": "F",
                "categoryType": "Application",
                "dropclass": "test.test_drop.SumupContainerChecksum",
                "node": ip_addr_2,
            },
        ]
        add_test_reprodata(g1)
        add_test_reprodata(g2)
        rels = [
            DROPRel("D", DROPLinkType.INPUT, "E"),
            DROPRel("D", DROPLinkType.INPUT, "F"),
        ]
        quickDeploy(dm1, sessionId, g1, {nm_conninfo(1): rels})
        quickDeploy(dm2, sessionId, g2, {nm_conninfo(0): rels})

        self.assertEqual(3, len(dm1._sessions[sessionId].drops))
        self.assertEqual(2, len(dm2._sessions[sessionId].drops))

        cons_nodes = dm1._sessions[sessionId].drops["D"].get_consumers_nodes()

        self.assertTrue(ip_addr_1 in cons_nodes)
        self.assertTrue(ip_addr_2 in cons_nodes)

        dm1.destroySession(sessionId)
        dm2.destroySession(sessionId)

    def test_run_streaming_consumer_remotely(self):
        """
        A test that checks that a streaming consumer works correctly across
        node managers when its input is in a different node, like this:

        DM #1                 DM #2
        ==================    ==============
        | A --> B --> C -|----|--> D --> E |
        ==================    ==============

        Here B is anormal application and D is a streaming consumer of C.
        We use A and E to compare that all data flows correctly.
        """

        g1 = [
            memory("A"),
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.CopyApp",
                "inputs": ["A"],
                "outputs": ["C"],
            },
            memory("C"),
        ]
        g2 = [
            {
                "oid": "D",
                "categoryType": "Application",
                "dropclass": "dlg.apps.crc.CRCStreamApp",
                "outputs": ["E"],
            },
            memory("E"),
        ]
        rels = [DROPRel("C", DROPLinkType.STREAMING_INPUT, "D")]
        a_data = os.urandom(32)
        e_data = str(crc32c(a_data, 0)).encode("utf8")
        self._test_runGraphInTwoNMs(g1, g2, rels, a_data, e_data, leaf_oid="E")

    def test_run_streaming_consumer_remotely2(self):
        """
        Like above, but C is hostd by DM #2.
        """

        g1 = [
            memory("A"),
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.CopyApp",
                "inputs": ["A"],
            },
        ]
        g2 = [
            memory("C"),
            {
                "oid": "D",
                "categoryType": "Application",
                "dropclass": "dlg.apps.crc.CRCStreamApp",
                "streamingInputs": ["C"],
                "outputs": ["E"],
            },
            memory("E"),
        ]
        rels = [DROPRel("C", DROPLinkType.OUTPUT, "B")]
        a_data = os.urandom(32)
        e_data = str(crc32c(a_data, 0)).encode("utf8")
        self._test_runGraphInTwoNMs(g1, g2, rels, a_data, e_data, leaf_oid="E")


class TestDMMultiThreading(NodeManagerTestsBase, unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_processes = False

    def test_run_invalid_shmem_graph(self):
        """
        Our shared memory implementation does not support Python < 3.7
        This test asserts that a graph containing shared memory drops will not run if running
        in python < 3.8, and that it *does* run with python >= 3.8
        """

        graph = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.SharedMemoryDROP",
            }
        ]
        graph = add_test_reprodata(graph)
        dm = self._start_dm()
        sessionID = "s1"
        if sys.version_info < (3, 8):
            self.assertRaises(
                NotImplementedError, quickDeploy, dm, sessionID, graph
            )
        else:
            quickDeploy(dm, sessionID, graph)
            self.assertEqual(1, len(dm._sessions[sessionID].drops))
            dm.destroySession(sessionID)

class TestDMMultiProcessing(NodeManagerTestsBase, unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_processes = True


import pyrap.tables as pt

oldtable = "/scratch/jason/chiles_original.ms"
newtable = "/scratch/jason/chiles_adios.ms"
t = pt.table(oldtable)
dmdef = t.getdminfo()

print("Original Table: **************************")
for i in dmdef:
    print(i)
    print(dmdef[i])

dmdef["*17"]["TYPE"] = "AdiosStMan"
print("New Table: ***********************")
for i in dmdef:
    print(i)
    print(dmdef[i])

t.copy(newtable, True, True, dminfo=dmdef)



#!/usr/bin/python
#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
This module replays the logs produced by the local monitor, which collects
graph status on supercomputers

this module depends on graphviz, which by default is not installed together
with Daliuge. To run this module, please
1. install graphviz
2. pip install pygraphviz

this module also depends on networkx (included in Daliuge), which produces the
edge list that becomes input for gephi vis tool.
"""
import filecmp
import json
import logging
import optparse
import os
import sqlite3 as dbdrv
import sys
from collections import defaultdict
from datetime import datetime as dt
from xml.etree.ElementTree import ElementTree

import commands
import networkx as nx
import numpy as np
import pygraphviz as pgv

logger = logging.getLogger(__name__)

ORIGINAL_COLOR = (87, 87, 87)
YELLOW_COLOR = (255, 255, 0)
# GREEN_COLOR = (102, 255, 178)
GREEN_COLOR = (0, 255, 0)
RED_COLOR = (255, 0, 0)
BLUE_COLOR = (102, 178, 255)

# TODO place java class in the current dir, and include it in Git repo
java_cmd = "java -classpath /tmp/classes:/Users/Chen/proj/gephi-toolkit/gephi-toolkit-0.9.1-all.jar dlg.deploy.utils.export_graph"

sql_create_status = """\
create table ac(ts integer, oid varchar(256), status integer);
create index ac_ts_index on ac(ts);
.separator ","
.import {0} ac
"""

sql_query = """\
SELECT max(ts), oid, status FROM ac
WHERE ts >= {0} and ts < {1}
GROUP BY oid
"""


class GraphPlayer(object):
    def __init__(self, graph_path, status_path):
        for fp in [graph_path, status_path]:
            if fp is None:
                raise Exception("JSON file is none!")
            if not os.path.exists(fp):
                raise Exception("JSON file not found: {0}".format(fp))
        self.oid_gnid_dict = dict()
        self.node_graph_dict = defaultdict(
            list
        )  # k - node-ip, v - a list of tuple of (graph node_id, downstream drop ids)
        self.gnid_ip_dict = dict()
        self.status_path = status_path
        with open(graph_path) as f:
            logger.info("Loading graph from file %s", graph_path)
            self.pg_spec = json.load(f)["g"]
        for i, dropspec in enumerate(self.pg_spec.values()):
            gnid = str(i)
            self.oid_gnid_dict[dropspec["oid"]] = gnid
            ip = dropspec["node"]
            self.node_graph_dict[ip].append(
                (gnid, self.get_downstream_drop_ids(dropspec))
            )
            self.gnid_ip_dict[gnid] = ip
        logger.info("oid to gid mapping done")

    def status_to_colour(self, st):
        st = int(st)
        if st == 0:  # INITIALIZED or NOT_RUN
            r, g, b = ORIGINAL_COLOR
        elif st == 1:  # WRITING or RUNNING
            # logger.debug("running")
            r, g, b = YELLOW_COLOR
        elif st == 2:  # COMPLETED or FINISHED
            # logger.debug("completed")
            r, g, b = GREEN_COLOR
        elif st == 3:  # ERROR
            # logger.debug("error")
            r, g, b = RED_COLOR
        else:
            r, g, b = BLUE_COLOR
        return (r, g, b)

    def convert_to_gexf_status(self, oid, status, colour_dict):
        """
        insert gexf colour line, e.g.
            <viz:color r="13" g="0" b="92"></viz:color>
        into the colour_dict using gnid as the key
        """
        gnid = self.oid_gnid_dict[oid]
        if "execStatus" in status:
            st = status["execStatus"]
        else:
            st = status["status"]
        r, g, b = self.status_to_colour(st)
        colour_dict[gnid] = '<viz:color r="%d" g="%d" b="%d"></viz:color>' % (
            r,
            g,
            b,
        )

    def parse_status(self, gexf_file, out_dir=None, remove_gexf=False):
        """
        Parse the graph status json file and
        produce gexf file for each status line
        """
        if gexf_file is None or (not os.path.exists(gexf_file)):
            raise Exception("GEXF file {0} not found".format(gexf_file))
        if out_dir is None:
            out_dir = os.path.dirname(gexf_file)
        with open(gexf_file) as gf:
            gexf_list = gf.readlines()
        logger.info("Gexf file '%s' loaded", gexf_file)
        with open(self.status_path) as f:
            for i, line in enumerate(f):
                colour_dict = dict()
                colour_dict[
                    "{}"
                ] = '<viz:color r="%d" g="%d" b="%d"></viz:color>\n' % (
                    ORIGINAL_COLOR[0],
                    ORIGINAL_COLOR[1],
                    ORIGINAL_COLOR[2],
                )
                output_lines = []
                pgs = json.loads(line)
                ts = pgs["ts"]
                gs = pgs["gs"]
                for oid, status in gs.iteritems():
                    self.convert_to_gexf_status(oid, status, colour_dict)
                node_id = None
                nodes_done = False
                for gexf_line in gexf_list:
                    if (not nodes_done) and gexf_line.find("<edges>") > -1:
                        nodes_done = True
                    if nodes_done:
                        output_lines.append(gexf_line)
                    else:
                        if gexf_line.find("<viz:color") > -1:
                            # logger.debug("Starting with viz color")
                            if node_id is None:
                                raise Exception("wrong order")
                            output_lines.append(colour_dict[node_id])
                        else:
                            if gexf_line.find("<node id=") > -1:
                                # <node id="38345" label="38345">
                                # 38345
                                node_id = gexf_line.split()[1].split("=")[1][
                                    1:-1
                                ]
                            output_lines.append(gexf_line)
                # each line generate a new file
                new_gexf = "{0}/{1}.gexf".format(out_dir, ts)
                with open(new_gexf, "w") as fo:
                    for new_line in output_lines:
                        # fo.write('{0}{1}'.format(new_line, os.linesep))
                        # fo.write('{0}{1}'.format(new_line, '\n'))
                        fo.write(new_line)
                logger.info("GEXF file '%s' is generated", new_gexf)
                new_png = new_gexf.split(".gexf")[0] + ".png"
                cmd = "{0} {1} {2}".format(java_cmd, new_gexf, new_png)
                ret = commands.getstatusoutput(cmd)
                if ret[0] != 0:
                    logger.error(
                        "Fail to print png from %s to %s: %s",
                        new_gexf,
                        new_png,
                        ret[1],
                    )
                del colour_dict
                if remove_gexf:
                    try:
                        os.remove(new_gexf)
                    except:
                        pass

    def get_downstream_drop_ids(self, dropspec):
        if dropspec["categoryType"] in ["Application", "dropclass"]:
            ds_kw = "outputs"  # down stream key word
        elif dropspec["categoryType"] in ["Data", "data"]:
            ds_kw = "consumers"
        else:
            ds_kw = "None"
        if ds_kw in dropspec:
            return dropspec[ds_kw]
        else:
            return []

    def build_node_graph(self):
        """
        A graph contains all compute nodes and their relationships
        """
        G = pgv.AGraph(strict=False, directed=True)
        temp_dict = defaultdict(int)  # key - from_to_ip, val - counter

        for i, ip in enumerate(self.node_graph_dict.keys()):
            G.add_node(ip, shape="rect", label="%d" % i)
        logger.info("All nodes added")

        for ip, droplist in self.node_graph_dict.iteritems():
            for gnid, dropids in droplist:
                for did in dropids:
                    tip = self.gnid_ip_dict[self.oid_gnid_dict[did]]
                    k = "{0}_{1}".format(ip, tip)
                    temp_dict[k] += 1

        for k, v in temp_dict.iteritems():
            ks = k.split("_")
            G.add_edge(ks[0], ks[1], weight=v)

        return G

    def get_state_changes(
        self, gexf_file, grep_log_file, steps=400, out_dir=None
    ):
        """
        grep -R "changed to state" --include=*.log . > statelog.txt
        the simulation time will be evenly divided up by "steps"
        """
        # convert grep_log_file to csv
        if out_dir is None:
            out_dir = os.path.dirname(gexf_file)
        # grep_log_file = '{0}/statelog.txt'.format(out_dir)
        # cmd = ""
        csv_file = "{0}/csv_file.csv".format(out_dir)
        sqlite_file = "{0}/sqlite_file.sqlite".format(out_dir)

        if not os.path.exists(csv_file):
            with open(grep_log_file, "r") as f:
                alllines = f.readlines()
            with open(csv_file, "w") as fo:
                for line in alllines:
                    ts = (
                        line.split("[DEBUG]")[0].split("dlgNM.log:")[1].strip()
                    )
                    ts = int(
                        dt.strptime(ts, "%Y-%m-%d %H:%M:%S,%f").strftime("%s")
                    )
                    oid = line.split("oid=")[1].split()[0]
                    state = line.split()[-1]
                    fo.write(
                        "{0},{1},{2},{3}".format(ts, oid, state, os.linesep)
                    )
        else:
            logger.info("csv file already exists: %s", csv_file)

        if not os.path.exists(sqlite_file):
            sql = sql_create_status.format(csv_file)
            sql_file = csv_file.replace(".csv", ".sql")
            with open(sql_file, "w") as fo:
                fo.write(sql)
            cmd = "sqlite3 {0} < {1}".format(sqlite_file, sql_file)
            ret = commands.getstatusoutput(cmd)
            if ret[0] != 0:
                logger.error("fail to create sqlite: %s", ret[1])
                return
        else:
            logger.info("sqlite file already exists: %s", sqlite_file)

        dbconn = dbdrv.connect(sqlite_file)
        q = "SELECT min(ts) from ac"
        cur = dbconn.cursor()
        cur.execute(q)
        lhs = cur.fetchall()[0][0]
        cur.close()

        q = "SELECT max(ts) from ac"
        cur = dbconn.cursor()
        cur.execute(q)
        rhs = cur.fetchall()[0][0]
        cur.close()

        step_size = (rhs - lhs) / steps
        if step_size == 0:
            return
        lr = list(np.arange(lhs, rhs, step_size))
        if lr[-1] != rhs:
            lr.append(rhs)

        last_gexf = gexf_file
        for i, el in enumerate(lr[0:-1]):
            a = el
            b = lr[i + 1]
            step_name = "{0}-{1}".format(a, b)
            logger.debug("stepname: %s", step_name)
            new_gexf = "{0}/{1}.gexf".format(out_dir, step_name)
            if os.path.exists(new_gexf):
                logger.info("%s already exists, ignore", new_gexf)
                last_gexf = new_gexf
                continue
            sql = sql_query.format(a, b)
            logger.debug(sql)
            cur = dbconn.cursor()
            cur.execute(sql)
            drops = cur.fetchall()
            cur.close()

            # build a dictionary
            drop_dict = dict()  # key - gnid, value: drop status
            for drop in drops:
                gnid = self.oid_gnid_dict[drop[1]]
                drop_dict[gnid] = drop[2]

            tree = ElementTree()
            tree.parse(last_gexf)
            root = tree.getroot()
            for node in root.iter("{http://www.gexf.net/1.3}node"):
                gnid = node.attrib["id"]
                if not gnid in drop_dict:
                    continue
                new_status = drop_dict[gnid]
                r, g, b = self.status_to_colour(new_status)
                colour = node[2]
                colour.attrib["r"] = "{0}".format(r)
                colour.attrib["g"] = "{0}".format(g)
                colour.attrib["b"] = "{0}".format(b)
            tree.write(new_gexf)
            logger.info("GEXF file '%s' is generated", new_gexf)
            del drop_dict
            if not filecmp.cmp(last_gexf, new_gexf, False):
                new_png = new_gexf.split(".gexf")[0] + ".png"
                cmd = "{0} {1} {2}".format(java_cmd, new_gexf, new_png)
                ret = commands.getstatusoutput(cmd)
            else:
                logger.info("Identical %s == %s", new_gexf, last_gexf)
            last_gexf = new_gexf
            if ret[0] != 0:
                logger.error(
                    "Fail to print png from %s to %s: %s",
                    last_gexf,
                    new_png,
                    ret[1],
                )

    def build_drop_subgraphs(self, node_range="[0:20]"):
        pass

    def build_drop_fullgraphs(self, do_subgraph=False, graph_lib="pygraphviz"):
        """
        this only works for small graphs, no way for # of drops > 1000
        """
        if "pygraphviz" == graph_lib:
            G = pgv.AGraph(strict=True, directed=True)
        else:
            G = nx.Graph()
            do_subgraph = False
        subgraph_dict = defaultdict(
            list
        )  # k - node-ip, v - a list of graph nodes
        oid_gnid_dict = dict()

        for i, oid in enumerate(self.pg_spec.keys()):
            oid_gnid_dict[oid] = str(i)
        logger.info("oid to gid mapping done")

        for dropspec in self.pg_spec.itervalues():
            gid = oid_gnid_dict[dropspec["oid"]]
            ip = dropspec["node"]
            subgraph_dict[ip].append(gid)
            if dropspec["categoryType"] in ["Application", "app"]:
                G.add_node(
                    gid, shape="rect", label=""
                )  # , fixedsize=True, hight=.05, width=.05)
            elif dropspec["categoryType"] in ["Data", "data"]:  # parallelogram
                G.add_node(
                    gid, shape="circle", label=""
                )  # , fixedsize=True, hight=.05, width=.05)
        logger.info("Graph nodes added")

        for dropspec in self.pg_spec.itervalues():
            gid = oid_gnid_dict[dropspec["oid"]]
            if dropspec["categoryType"] in ["Application", "app"]:
                ds_kw = "outputs"  # down stream key word
            elif dropspec["categoryType"] in ["Data", "data"]:
                ds_kw = "consumers"
            else:
                ds_kw = "None"
            if ds_kw in dropspec:
                for doid in dropspec[ds_kw]:
                    G.add_edge(gid, oid_gnid_dict[doid])
        logger.info("Graph edges added")

        if do_subgraph:
            for i, subgraph_nodes in enumerate(subgraph_dict.values()):
                # we don't care about the subgraph label or rank
                subgraph = G.add_subgraph(
                    subgraph_nodes,
                    label="%d" % i,
                    name="cluster_%d" % i,
                    rank="same",
                )
                subgraph.graph_attr["rank"] = "same"
            logger.info("Subgraph added")

        return G


if __name__ == "__main__":
    """
    1. create edge list from 'monitor_g.log'
    2. run gephi manully to (1) finalise the layout, and
                            (2) store the layout in the gexf file
    3. alternatively, one can use existing gexf file from previous runs

    e.g. python monitor_replayer.py -x /Users/Chen/Documents/FromWeb/Galaxy/2016-06-30/galaxy.gexf
    -r /Users/Chen/Documents/FromWeb/Galaxy/2016-07-15/logs/2016-07-15_11-59-21/statelog.txt
    -u /Users/Chen/Documents/FromWeb/Galaxy/2016-07-15/gen -o rrr -d eee
    -l /Users/Chen/Documents/FromWeb/Galaxy/2016-07-15/vis.log
    -g /Users/Chen/Documents/FromWeb/Galaxy/2016-07-15/logs/2016-07-15_11-59-21/0/monitor_g.log
    -t /Users/Chen/Documents/FromWeb/Galaxy/2016-07-15/logs/2016-07-15_11-59-21/0/monitor_g.log
    """
    parser = optparse.OptionParser()
    parser.add_option(
        "-g",
        "--graph_path",
        action="store",
        type="string",
        dest="graph_path",
        help="path to physical graph specification",
        default=None,
    )
    parser.add_option(
        "-t",
        "--status_path",
        action="store",
        type="string",
        dest="status_path",
        help="path to physical graph status",
        default=None,
    )
    parser.add_option(
        "-x",
        "--gexf_file",
        action="store",
        type="string",
        dest="gexf_file",
        help="path to gexf file produced from gephi",
        default=None,
    )
    parser.add_option(
        "-u",
        "--gexf_output_dir",
        action="store",
        type="string",
        dest="gexf_output_dir",
        help="path to gexf output directory",
        default=None,
    )
    parser.add_option(
        "-l",
        "--log_file",
        action="store",
        type="string",
        dest="log_file",
        help="logfile path",
        default=None,
    )
    parser.add_option(
        "-o",
        "--output_file",
        action="store",
        type="string",
        dest="output_file",
        help="output image file",
        default=None,
    )
    parser.add_option(
        "-d",
        "--dot_file",
        action="store",
        type="string",
        dest="dot_file",
        help="output do file",
        default=None,
    )
    parser.add_option(
        "-s",
        "--subgraph",
        action="store_true",
        dest="subgraph",
        help="create subgraph per node",
        default=False,
    )
    parser.add_option(
        "-e",
        "--edgelist",
        action="store_true",
        dest="edgelist",
        help="store edge list instead of dot file",
        default=False,
    )

    parser.add_option(
        "-r",
        "--grep_log_file",
        action="store",
        type="string",
        dest="grep_log_file",
        help="grep log file",
        default=None,
    )

    (options, args) = parser.parse_args()

    if None == options.log_file or None == options.graph_path:
        parser.print_help()
        sys.exit(1)

    FORMAT = "%(asctime)-15s [%(levelname)5.5s] [%(threadName)15.15s] %(name)s#%(funcName)s:%(lineno)s %(message)s"
    logging.basicConfig(
        filename=options.log_file, level=logging.DEBUG, format=FORMAT
    )

    if options.edgelist and options.dot_file is not None:
        logger.info("Loading networx graph from file %s", options.graph_path)
        gp = GraphPlayer(options.graph_path, options.status_path)
        g = gp.build_drop_fullgraphs(graph_lib="networkx")
        nx.write_edgelist(g, options.dot_file)
        sys.exit(0)

    if None == options.output_file or None == options.dot_file:
        parser.print_help()
        sys.exit(1)

    gp = GraphPlayer(options.graph_path, options.status_path)
    # gp.parse_status(options.gexf_file, out_dir=options.gexf_output_dir)
    gp.get_state_changes(
        options.gexf_file,
        options.grep_log_file,
        out_dir=options.gexf_output_dir,
    )
    """
    g = gp.build_node_graph()
    #g = gp.build_drop_fullgraphs(options.graph_path, do_subgraph=options.subgraph)
    g.write(options.dot_file)
    logger.info("Graph viz obj created")
    g.layout(prog='sfdp', args='-Goverlap=prism -Gsize=67!')
    #g.layout(prog='sfdp')
    #g.layout(prog='dot')
    logger.info("Graph viz layout computed")
    g.draw(options.output_file)
    logger.info("Graph viz file drawn to: {0}".format(options.output_file))
    """



from abc import ABC, abstractmethod
from collections import OrderedDict
from concurrent.futures import Future
from typing import List, Callable
import logging
import math
import threading

from dlg.drop_loaders import load_pickle
from dlg.data.drops.container import ContainerDROP
from dlg.data.drops.data_base import DataDROP
from dlg.ddap_protocol import (
    AppDROPStates,
    DROPLinkType,
    DROPStates,
    DROPRel,
)
from dlg.utils import object_tracking
from dlg.exceptions import InvalidDropException, InvalidRelationshipException

from dlg.meta import (
    dlg_int_param,
)

logger = logging.getLogger(__name__)

track_current_drop = object_tracking("drop")


class DropRunner(ABC):
    """An executor for `run()`-ing an AppDROP"""

    @abstractmethod
    def run_drop(self, app_drop: "AppDROP") -> Future:
        """Executes `app_drop.run()`, returning a future with the result."""
        raise NotImplementedError


class SyncDropRunner(DropRunner):
    """
    A simple pool-like object that creates a new thread for each invocation.
    """

    def run_drop(self, app_drop: "AppDROP") -> Future:
        """Run drop synchronously."""
        future = Future()

        try:
            res = app_drop.run()
            future.set_result(res)
        except BaseException as e:
            future.set_exception(e)

        return future


def run_on_daemon_thread(func: Callable, *args, **kwargs) -> Future:
    """Runs a callable on a daemon thread, meaning it will be
    ungracefully terminated if the process ends."""
    future = Future()

    def thread_target():
        try:
            res = func(*args, **kwargs)
            future.set_result(res)
        except BaseException as e:
            future.set_exception(e)

    t = threading.Thread(target=thread_target)
    t.daemon = True
    t.start()

    return future


_SYNC_DROP_RUNNER = SyncDropRunner()


# ===============================================================================
# AppDROP classes follow
# ===============================================================================


class AppDROP(ContainerDROP):
    """
    An AppDROP is a DROP representing an application that reads data
    from one or more DataDROPs (its inputs), and writes data onto one or more
    DataDROPs (its outputs).

    AppDROPs accept two different kind of inputs: "normal" and "streaming"
    inputs. Normal inputs are DataDROPs that must be on the COMPLETED state
    (and therefore their data must be fully written) before this application is
    run, while streaming inputs are DataDROPs that feed chunks of data into
    this application as the data gets written into them.

    This class contains two methods that need to be overwritten by
    subclasses: `dropCompleted`, invoked when input DataDROPs move to
    COMPLETED, and `dataWritten`, invoked with the data coming from streaming
    inputs.

    How and when applications are executed is completely up to the app component
    developer, and is not enforced by this base class. Some applications might need
    to be run at `initialize` time, while other might start during the first invocation
    of `dataWritten`. A common scenario anyway is to start an application only
    after all its inputs have moved to COMPLETED (implying that none of them is
    an streaming input); for these cases see the `BarrierAppDROP`.
    """

    def __getstate__(self):
        state = super().__getstate__()
        del state["_drop_runner"]
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._drop_runner = _SYNC_DROP_RUNNER

    def initialize(self, **kwargs):
        super(AppDROP, self).initialize(**kwargs)

        # Inputs and Outputs are the DROPs that get read from and written
        # to by this AppDROP, respectively. An input DROP will see
        # this AppDROP as one of its consumers, while an output DROP
        # will see this AppDROP as one of its producers.
        #
        # Input and output objects are later referenced by their *index*
        # (relative to the order in which they were added to this object)
        # Therefore we use an ordered dict to keep the insertion order.
        self._inputs = OrderedDict()
        self._outputs = OrderedDict()

        # Same as above, only that these correspond to the 'streaming' version
        # of the consumers
        self._streamingInputs = OrderedDict()

        # An AppDROP has a second, separate state machine indicating its
        # execution status.
        self._execStatus = AppDROPStates.NOT_RUN

        # by default run drops synchronously
        self._drop_runner: DropRunner = _SYNC_DROP_RUNNER

    @track_current_drop
    def addInput(self, inputDrop, back=True):
        uid = inputDrop.uid
        if uid not in self._inputs:
            self._inputs[uid] = inputDrop
            if back:
                inputDrop.addConsumer(self, False)

    @property
    def inputs(self) -> List[DataDROP]:
        """
        The list of inputs set into this AppDROP
        """
        return list(self._inputs.values())

    @track_current_drop
    def addOutput(self, outputDrop: DataDROP, back=True):
        if outputDrop is self:
            raise InvalidRelationshipException(
                DROPRel(outputDrop, DROPLinkType.OUTPUT, self),
                "Cannot add an AppConsumer as its own output",
            )
        uid = outputDrop.uid
        if uid not in self._outputs:
            self._outputs[uid] = outputDrop

            if back:
                outputDrop.addProducer(self, False)

            # Subscribe the output DROP to events sent by this AppDROP when it
            # finishes its execution.
            self.subscribe(outputDrop, "producerFinished")

    @property
    def outputs(self) -> List[DataDROP]:
        """
        The list of outputs set into this AppDROP
        """
        return list(self._outputs.values())

    def addStreamingInput(self, streamingInputDrop, back=True):
        if streamingInputDrop not in self._streamingInputs.values():
            uid = streamingInputDrop.uid
            self._streamingInputs[uid] = streamingInputDrop
            if back:
                streamingInputDrop.addStreamingConsumer(self, False)

    @property
    def streamingInputs(self) -> List[DataDROP]:
        """
        The list of streaming inputs set into this AppDROP
        """
        return list(self._streamingInputs.values())

    def _generateNamedPorts(self, ports):
        """
        Generates a named mapping of ports to data drops. Can only be called during run().
        """
        port_type = {"inputs": "InputPort", "outputs": "OutputPort"}
        named_ports: OrderedDict[str, DataDROP] = OrderedDict()
        port_dict = self.__getattribute__(f"_{ports}")
        if (
            ports in self.parameters
            and len(self.parameters[ports]) > 0
            and isinstance(self.parameters[ports][0], dict)
        ):
            for i in range(len(port_dict)):
                key = list(self.parameters[ports][i].values())[0]
                value = port_dict[list(self.parameters[ports][i].keys())[0]]
                if key not in named_ports:
                    named_ports[key] = value
                else:
                    if isinstance(named_ports[key], list):
                        named_ports[key].append(value)
                    else:
                        named_ports[key] = [named_ports[key], value]
        elif (
            ports in self.parameters
            and len(self.parameters[ports]) > 0
            and isinstance(self.parameters[ports], list)
        ):
            # This enablkes the gather to work
            return {}
        elif "applicationArgs" in self.parameters:
            for key, field in self.parameters["applicationArgs"].items():
                if field["usage"] in [port_type[ports], "InputOutput"]:
                    named_ports[field["name"]] = field
        return named_ports

    def handleEvent(self, e):
        """
        Handles the arrival of a new event. Events are delivered from those
        objects this DROP is subscribed to.
        """
        if e.type == "dropCompleted":
            self.dropCompleted(e.uid, e.status)

    def dropCompleted(self, uid, drop_state):
        """
        Callback invoked when the DROP with UID `uid` (which is either a
        normal or a streaming input of this AppDROP) has moved to the
        COMPLETED or ERROR state. By default no action is performed.
        """

    def dataWritten(self, uid, data):
        """
        Callback invoked when `data` has been written into the DROP with
        UID `uid` (which is one of the streaming inputs of this AppDROP).
        By default no action is performed
        """

    @property
    def execStatus(self):
        """
        The execution status of this AppDROP
        """
        return self._execStatus

    @execStatus.setter
    def execStatus(self, execStatus):
        if self._execStatus == execStatus:
            return
        self._execStatus = execStatus
        self._fire("execStatus", execStatus=execStatus)

    def _notifyAppIsFinished(self):
        """
        Method invoked by subclasses when the execution of the application is
        over. Subclasses must make sure that both the status and execStatus
        properties are set to their correct values correctly before invoking
        this method.
        """
        is_error = self._execStatus == AppDROPStates.ERROR
        if is_error:
            self.status = DROPStates.ERROR
        else:
            self.status = DROPStates.COMPLETED
        logger.debug(
            "Moving %r to %s",
            self.oid,
            "FINISHED" if not is_error else "ERROR",
        )
        self._fire("producerFinished", status=self.status, execStatus=self.execStatus)
        self.completedrop()

    def cancel(self):
        """Moves this application drop to its CANCELLED state"""
        super(AppDROP, self).cancel()
        self.execStatus = AppDROPStates.CANCELLED

    def skip(self):
        """Moves this application drop to its SKIPPED state"""
        super().skip()

        prev_execStatus = self.execStatus
        self.execStatus = AppDROPStates.SKIPPED
        for o in self._outputs.values():
            o.skip()

        logger.debug(f"Moving {self.__repr__()} to SKIPPED")
        if prev_execStatus in [AppDROPStates.NOT_RUN]:
            self._fire(
                "producerFinished",
                status=self.status,
                execStatus=self.execStatus,
            )


class InputFiredAppDROP(AppDROP):
    """
    An InputFiredAppDROP accepts no streaming inputs and waits until a given
    amount of inputs (called *effective inputs*) have moved to COMPLETED to
    execute its 'run' method, which must be overwritten by subclasses. This way,
    this application allows to continue the execution of the graph given a
    minimum amount of inputs being ready. The transitions of subsequent inputs
    to the COMPLETED state have no effect.

    Normally only one call to the `run` method will happen per application.
    However users can override this by specifying a different number of tries
    before finally giving up.

    The amount of effective inputs must be less or equal to the amount of inputs
    added to this application once the graph is being executed. The special
    value of -1 means that all inputs are considered as effective, in which case
    this class acts as a BarrierAppDROP, effectively blocking until all its
    inputs have moved to the COMPLETED, SKIPPED or ERROR state. Setting this
    value to anything other than -1 or the number of inputs, results in
    late arriving inputs to be ignored, even if they would successfully finish.
    This requires careful implementation of the upstream and downstream apps to
    deal with this situation. It is only really useful to control a combination
    of maximum allowed execution time and acceptable number of completed inputs.

    An input error threshold controls the behavior of the application given an
    error in one or more of its inputs (i.e., a DROP moving to the ERROR state).
    The threshold is a value within 0 and 100 that indicates the tolerance
    to erroneous effective inputs, and after which the application will not be
    run but moved to the ERROR state itself instead.
    """

    input_error_threshold = dlg_int_param("Input error threshold (0 and 100)", 0)
    n_effective_inputs = dlg_int_param("Number of effective inputs", -1)
    n_tries = dlg_int_param("Number of tries", 1)

    def initialize(self, **kwargs):
        super(InputFiredAppDROP, self).initialize(**kwargs)
        if "_dlg_session_id" in kwargs:
            self._dlg_session_id = kwargs["_dlg_session_id"]
        self._completedInputs = []
        self._errorInputs = []
        self._skippedInputs = []

        # Error threshold must be within 0 and 100
        if self.input_error_threshold < 0 or self.input_error_threshold > 100:
            raise InvalidDropException(
                self, "%r: input_error_threshold not within [0,100]" % (self,)
            )

        # Amount of effective inputs
        if "n_effective_inputs" not in kwargs:
            raise InvalidDropException(
                self, "%r: n_effective_inputs is mandatory" % (self,)
            )

        if self.n_effective_inputs < -1 or self.n_effective_inputs == 0:
            raise InvalidDropException(
                self,
                "%r: n_effective_inputs must be > 0 or equals to -1" % (self,),
            )

        # Number of tries
        if self.n_tries < 1:
            raise InvalidDropException(
                self, "Invalid n_tries, must be a positive number"
            )

    def addStreamingInput(self, streamingInputDrop, back=True):
        raise InvalidRelationshipException(
            DROPRel(streamingInputDrop, DROPLinkType.STREAMING_INPUT, self),
            "InputFiredAppDROPs don't accept streaming inputs",
        )

    def dropCompleted(self, uid, drop_state):
        super(InputFiredAppDROP, self).dropCompleted(uid, drop_state)

        logger.debug(
            "Received notification from input drop: uid=%s, state=%d",
            uid,
            drop_state,
        )

        # A value of -1 means all inputs
        n_inputs = len(self._inputs)
        n_eff_inputs = self.n_effective_inputs
        if n_eff_inputs == -1:
            n_eff_inputs = n_inputs

        # More effective inputs than inputs, this is a horror
        if n_eff_inputs > n_inputs:
            raise Exception(
                "%r: More effective inputs (%d) than inputs (%d)"
                % (self, self.n_effective_inputs, n_inputs)
            )

        if drop_state == DROPStates.ERROR:
            self._errorInputs.append(uid)
        elif drop_state == DROPStates.COMPLETED:
            self._completedInputs.append(uid)
        elif drop_state == DROPStates.SKIPPED:
            self._skippedInputs.append(uid)
        else:
            raise Exception("Invalid DROP state in dropCompleted: %s" % drop_state)

        error_len = len(self._errorInputs)
        ok_len = len(self._completedInputs)
        skipped_len = len(self._skippedInputs)

        # We have enough inputs to proceed
        if (skipped_len + error_len + ok_len) == n_eff_inputs:
            # calculate the number of errors that have already occurred
            percent_failed = math.floor((error_len / float(n_eff_inputs)) * 100)
            if percent_failed > 0:
                logger.debug(
                    "Error rate on inputs for %r: %d/%d",
                    self,
                    percent_failed,
                    self.input_error_threshold,
                )

            # if we hit the input error threshold then ERROR the drop and move on
            if percent_failed > self.input_error_threshold:
                logger.info(
                    "Error threshold reached on %r, not executing it: %d/%d",
                    self,
                    percent_failed,
                    self.input_error_threshold,
                )

                self.execStatus = AppDROPStates.ERROR
                self.status = DROPStates.ERROR
                self._notifyAppIsFinished()
            elif skipped_len == n_eff_inputs:
                self.skip()
            else:
                self.async_execute()

    def async_execute(self):
        # TODO Do we need another thread pool for this?
        # Careful, trying to run this on the same threadpool as the
        # DropRunner can cause deadlocks
        return run_on_daemon_thread(self._execute_and_log_exception)

    def _execute_and_log_exception(self):
        try:
            self.execute()
        except:
            logger.exception("Unexpected exception during drop (%r) execution", self)

    @track_current_drop
    def execute(self, _send_notifications=True):
        """
        Manually trigger the execution of this application.

        This method is normally invoked internally when the application detects
        all its inputs are COMPLETED.
        """

        # TODO: We need to be defined more clearly how the state is set in
        #       applications, for the time being they follow their execState.

        # Run at most self._n_tries if there are errors during the execution
        logger.debug("Executing %r", self.oid)
        tries = 0
        drop_state = DROPStates.COMPLETED
        self.execStatus = AppDROPStates.RUNNING
        while tries < self.n_tries:
            try:
                fut = self._drop_runner.run_drop(self)
                fut.result()

                if self.execStatus == AppDROPStates.CANCELLED:
                    return
                self.execStatus = AppDROPStates.FINISHED
                break
            except:
                if self.execStatus == AppDROPStates.CANCELLED:
                    return
                tries += 1
                logger.exception(
                    "Error while executing %r (try %s/%s)",
                    self,
                    tries,
                    self.n_tries,
                )

        # We gave up running the application, go to error
        if tries == self.n_tries:
            self.execStatus = AppDROPStates.ERROR
            drop_state = DROPStates.ERROR

        self.status = drop_state
        if _send_notifications:
            self._notifyAppIsFinished()

    def _run(self):
        """
        Run this application. It can be safely assumed that at this point all
        the required inputs are COMPLETED.

        This will first set the named input params and then call the run method
        provided by the implementation.
        """
        named_inputs = self._generateNamedPorts("inputs")
        logger.debug("named inputs identified: %s", named_inputs)
        for attr_name in named_inputs:
            # if not isinstance(named_inputs[attr_name], list):
            #     self.__setattr__(
            #         attr_name, load_pickle(named_inputs[attr_name])
            #     )
            # else:
            self.__setattr__(attr_name, named_inputs[attr_name])

        named_outputs = self._generateNamedPorts("outputs")
        logger.debug("named outputs identified: %s", named_outputs)
        for attr_name in named_outputs:
            if not isinstance(named_outputs[attr_name], list):
                self.__setattr__(attr_name, named_outputs[attr_name])
        # if "run" in dir(self):  # we might not have implemented the run method
        #     self.run()

    # TODO: another thing we need to check
    def exists(self):
        return True


class BarrierAppDROP(InputFiredAppDROP):
    """
    A BarrierAppDROP is an InputFireAppDROP that waits for all its inputs to
    complete, effectively blocking the flow of the graph execution.
    """

    def initialize(self, **kwargs):
        # Blindly override existing value if any
        kwargs["n_effective_inputs"] = -1
        super().initialize(**kwargs)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import base64
import io
import json
import os
import pickle
import random
import string
import sys

from dlg.common.reproducibility.reproducibility import common_hash
from dlg.data.drops.data_base import DataDROP, logger
from dlg.data.io import SharedMemoryIO, MemoryIO


def parse_pydata(pd_dict: dict) -> bytes:
    """
    Parse and evaluate the pydata argument to populate memory during initialization

    :param pd_dict: the pydata dictionary from the graph node

    :returns a byte encoded value
    """
    pydata = pd_dict["value"]
    logger.debug(f"pydata value provided: {pydata}, {type(pydata)}")

    if pd_dict["type"].lower() == "json":
        try:
            pydata = json.loads(pydata)
        except:
            pydata = pydata.encode()
    elif pd_dict["type"].lower() == "int":
        try:
            pydata = int(pydata)
        except:
            pydata = pydata.encode()
    elif pd_dict["type"].lower() == "float":
        try:
            pydata = float(pydata)
        except:
            pydata = pydata.encode()
    elif pd_dict["type"].lower() == "boolean":
        try:
            pydata = bool(pydata)
        except:
            pydata = pydata.encode()
    elif pd_dict["type"].lower() == "object":
        pydata = base64.b64decode(pydata.encode())
        try:
            pydata = pickle.loads(pydata)
        except:
            raise
    return pickle.dumps(pydata)


##
# @brief Memory
# @details In-memory storage of intermediate data products
# @par EAGLE_START
# @param category Memory
# @param tag daliuge
# @param pydata None/String/ApplicationArgument/NoPort/ReadWrite//False/False/Data to be loaded into memory
# @param dummy /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Dummy port
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param dropclass dlg.data.drops.memory.InMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @par EAGLE_END
class InMemoryDROP(DataDROP):
    """
    A DROP that points data stored in memory.
    """

    # Allow in-memory drops to be automatically removed by default
    def __init__(self, *args, **kwargs):
        if "persist" not in kwargs:
            kwargs["persist"] = False
        if "expireAfterUse" not in kwargs:
            kwargs["expireAfterUse"] = True
        if kwargs["persist"]:
            kwargs["expireAfterUse"] = False
        super().__init__(*args, **kwargs)

    def initialize(self, **kwargs):
        """
        If there is a pydata argument use that to populate the DROP
        """
        args = []
        pydata = None
        field_names = (
            [f["name"] for f in kwargs["fields"]] if "fields" in kwargs else []
        )
        if "pydata" in kwargs and not (
            "fields" in kwargs and "pydata" in field_names
        ):  # means that is was passed directly
            pydata = kwargs.pop("pydata")
            logger.debug("pydata value provided: %s, %s", pydata, kwargs)
            try:  # test whether given value is valid
                _ = pickle.loads(base64.b64decode(pydata))
                pydata = base64.b64decode(pydata)
            except:
                pydata = None
        elif "fields" in kwargs and "pydata" in field_names:
            data_pos = field_names.index("pydata")
            pydata = parse_pydata(kwargs["fields"][data_pos])
        args.append(pydata)
        logger.debug("Loaded into memory: %s", pydata)
        self._buf = io.BytesIO(*args)
        self.size = len(pydata) if pydata else 0

    def getIO(self):
        if (
            hasattr(self, "_tp")
            and hasattr(self, "_sessionId")
            and sys.version_info >= (3, 8)
        ):
            return SharedMemoryIO(self.oid, self._sessionId)
        else:
            return MemoryIO(self._buf)

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return "mem://%s/%d/%d" % (hostname, os.getpid(), id(self._buf))

    # Override
    def generate_reproduce_data(self):
        from dlg.droputils import allDropContents

        data = b""
        try:
            data = allDropContents(self, self.size)
        except Exception:
            logger.debug("Could not read drop reproduce data")
        return {"data_hash": common_hash(data)}


##
# @brief PythonObject
# @details A Python object stored in memory
# @par EAGLE_START
# @param category PythonObject
# @param tag daliuge
# @param object /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/object
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Object should be serialized
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in the object
# @param dropclass dlg.data.drops.memory.InMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @par EAGLE_END
class PythonObjectDROP(InMemoryDROP):
    """
    A python object is an InMemoryDROP but has a special category of PythonObject.

    This is only here to force the creation of the component for the palette.
    """

    pass


##
# @brief SharedMemory
# @details Data stored in shared memory
# @par EAGLE_START
# @param category SharedMemory
# @param tag daliuge
# @param pydata None/String/ApplicationArgument/NoPort/ReadWrite//False/False/Data to be loaded into memory
# @param dummy /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Dummy port
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param dropclass dlg.data.drops.memory.SharedMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @par EAGLE_END
class SharedMemoryDROP(DataDROP):
    """
    A DROP that points to data stored in shared memory.
    This drop is functionality equivalent to an InMemory drop running in a concurrent environment.
    In this case however, the requirement for shared memory is explicit.

    @WARNING Currently implemented as writing to shmem and there is no backup behaviour.
    """

    def initialize(self, **kwargs):
        args = []
        pydata = None
        if "pydata" in kwargs and not (
            "nodeAttributes" in kwargs and "pydata" in kwargs["nodeAttributes"]
        ):  # means that is was passed directly
            pydata = kwargs.pop("pydata")
            logger.debug("pydata value provided: %s", pydata)
            try:  # test whether given value is valid
                _ = pickle.loads(base64.b64decode(pydata.encode("latin1")))
                pydata = base64.b64decode(pydata.encode("latin1"))
            except:
                pydata = None
        elif "nodeAttributes" in kwargs and "pydata" in kwargs["nodeAttributes"]:
            pydata = parse_pydata(kwargs["nodeAttributes"]["pydata"])
        args.append(pydata)
        self._buf = io.BytesIO(*args)

    def getIO(self):
        if sys.version_info >= (3, 8):
            if hasattr(self, "_sessionId"):
                return SharedMemoryIO(self.oid, self._sessionId)
            else:
                # Using Drop without manager, just generate a random name.
                sess_id = "".join(
                    random.choices(string.ascii_uppercase + string.digits, k=10)
                )
                return SharedMemoryIO(self.oid, sess_id)
        else:
            raise NotImplementedError(
                "Shared memory is only available with Python >= 3.8"
            )

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return f"shmem://{hostname}/{os.getpid()}/{id(self._buf)}"



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import base64
import logging
import os
import pickle
import random
import unittest
import numpy

from dlg import droputils, drop_loaders
from dlg.apps import pyfunc
from dlg.apps.simple_functions import string2json
from dlg.ddap_protocol import DROPStates, DROPRel, DROPLinkType
from dlg.data.drops.memory import InMemoryDROP
from dlg.droputils import DROPWaiterCtx
from dlg.exceptions import InvalidDropException

from ..manager import test_dm

logger = logging.getLogger(__name__)


def func1(arg1):
    return arg1


def func2(arg1):
    return arg1 * 2


def func3():
    return ["b", "c", "d"]


def func_with_defaults(a, b=10, c=20, x=30, y=40, z=50):
    """Returns a - b * c + (y - x) * z. Default is a + 300"""
    res = a - b * c + (y - x) * z
    logger.info("%r - %r * %r + (%r - %r) * %r = %r", a, b, c, y, x, z, res)
    return res


def sum_with_args_and_kwarg(a, *args, **kwargs):
    """Returns a + kwargs['b'], or only a if no 'b' is found in kwargs"""
    b = kwargs.pop("b", 0)
    return a + sum(args) + b


def _PyFuncApp(oid, uid, f, **kwargs):
    fname = None
    if isinstance(f, str):
        fname = f = "test.apps.test_pyfunc." + f
    fw_kwargs = {
        k: v for k, v in kwargs.items() if k in ["input_parser", "output_parser"]
    }
    input_kws = [
        {k: v} for k, v in kwargs.items() if k not in ["input_parser", "output_parser"]
    ]
    fcode, fdefaults = pyfunc.serialize_func(f)
    return pyfunc.PyFuncApp(
        oid,
        uid,
        func_name=fname,
        func_code=fcode,
        func_defaults=fdefaults,
        inputs=input_kws,
        **fw_kwargs,
    )


class TestPyFuncApp(unittest.TestCase):
    def test_missing_function_param(self):
        self.assertRaises(InvalidDropException, pyfunc.PyFuncApp, "a", "a")

    def test_invalid_function_param(self):
        # The function doesn't have a module
        self.assertRaises(
            InvalidDropException, pyfunc.PyFuncApp, "a", "a", func_name="func1"
        )

    def test_function_invalid_module(self):
        # The function lives in an unknown module/package
        self.assertRaises(
            InvalidDropException,
            pyfunc.PyFuncApp,
            "a",
            "a",
            func_name="doesnt_exist.func1",
        )

    def test_function_builtin_module(self):
        # The function lives in an unknown module/package
        pyfunc.PyFuncApp("a", "a", func_name="object.__init__")

    def test_function_invalid_fname(self):
        # The function lives in an unknown module/package
        self.assertRaises(
            InvalidDropException,
            pyfunc.PyFuncApp,
            "a",
            "a",
            func_name="test.apps.test_pyfunc.doesnt_exist",
        )

    def test_valid_creation(self):
        _PyFuncApp("a", "a", "func1")

    def test_creation_with_code(self):
        def inner_function(x, y):
            return x + y

        _PyFuncApp("a", "a", inner_function)

    def test_pickle_func(self, f=lambda x: x, input_data="hello", output_data="hello"):
        a = InMemoryDROP("a", "a")
        b = _PyFuncApp("b", "b", f)
        c = InMemoryDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        with DROPWaiterCtx(self, c, 5):
            drop_loaders.save_pickle(a, input_data)
            a.setCompleted()
        for drop in a, b, c:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        self.assertEqual(output_data, drop_loaders.load_pickle(c))

    def test_eval_func(self, f=lambda x: x, input_data=None, output_data=None):
        input_data = [2, 2] if input_data is None else input_data
        output_data = [2, 2] if output_data is None else output_data

        a = InMemoryDROP("a", "a")
        b = _PyFuncApp(
            "b",
            "b",
            f,
            input_parser=pyfunc.DropParser.EVAL,
            output_parser=pyfunc.DropParser.EVAL,
        )
        c = InMemoryDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        with DROPWaiterCtx(self, c, 5):
            a.write(repr(input_data).encode("utf-8"))
            a.setCompleted()
        for drop in a, b, c:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        self.assertEqual(
            output_data,
            eval(droputils.allDropContents(c).decode("utf-8"), {}, {}),
        )

    def test_string2json_func(self, f=string2json, input_data=None, output_data=None):
        input_data = '["a", "b", "c"]' if input_data is None else input_data
        output_data = ["a", "b", "c"] if output_data is None else output_data

        a = InMemoryDROP("a", "a")
        b = _PyFuncApp(
            "b",
            "b",
            f,
        )
        c = InMemoryDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        with DROPWaiterCtx(self, c, 5):
            drop_loaders.save_pickle(a, input_data)
            a.setCompleted()
        for drop in a, b, c:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        numpy.testing.assert_equal(output_data, drop_loaders.load_pickle(c))

    def test_npy_func(self, f=lambda x: x, input_data=None, output_data=None):
        input_data = numpy.ones([2, 2]) if input_data is None else input_data
        output_data = numpy.ones([2, 2]) if output_data is None else output_data

        a = InMemoryDROP("a", "a")
        b = _PyFuncApp(
            "b",
            "b",
            f,
            input_parser=pyfunc.DropParser.NPY,
            output_parser=pyfunc.DropParser.NPY,
        )
        c = InMemoryDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        with DROPWaiterCtx(self, c, 5):
            drop_loaders.save_npy(a, input_data)
            a.setCompleted()
        for drop in a, b, c:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        numpy.testing.assert_equal(output_data, drop_loaders.load_npy(c))

    def _test_simple_functions(self, f, input_data, output_data):
        a, c = [InMemoryDROP(x, x) for x in ("a", "c")]
        b = _PyFuncApp("b", "b", f)
        b.addInput(a)
        b.addOutput(c)

        with DROPWaiterCtx(self, c, 5):
            a.write(pickle.dumps(input_data))
            a.setCompleted()

        for drop in a, b, c:
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        self.assertEqual(output_data, pickle.loads(droputils.allDropContents(c)))

    def test_func1(self):
        """Checks that func1 in this module works when wrapped"""
        data = os.urandom(64)
        self._test_simple_functions("func1", data, data)

    def test_func2(self):
        """Checks that func2 in this module works when wrapped"""
        n = random.randint(0, 1_000_000)
        self._test_simple_functions("func2", n, 2 * n)

    def test_inner_func(self):
        n = random.randint(0, 1_000_000)

        def f(x):
            return x + 2

        self._test_simple_functions(f, n, n + 2)

    def test_lambda(self):
        n = random.randint(0, 1_000_000)
        self._test_simple_functions(lambda x: x / 2, n, n / 2)

    def test_inner_func_with_closure(self):
        n = random.randint(0, 1_000_000)

        def f(x):
            return x + n

        self._test_simple_functions(f, n, n + n)

    def test_lambda_with_closure(self):
        n = random.randint(0, 1_000_000)
        self._test_simple_functions(lambda x: (x + n) / 2, n, (n + n) / 2)

    def _test_func3(self, output_drops, expected_outputs):
        a = _PyFuncApp("a", "a", "func3")
        for drop in output_drops:
            a.addOutput(drop)

        drops = [a] + droputils.listify(output_drops)
        a.execute()

        for drop in drops:
            self.assertEqual(DROPStates.COMPLETED, drop.status)

        for expected_output, drop in zip(expected_outputs, output_drops):
            self.assertEqual(
                expected_output, pickle.loads(droputils.allDropContents(drop))
            )  # @UndefinedVariable

    def test_func3_singleoutput(self):
        """
        Checks that func3 in this module works when wrapped as an application
        with multiple outputs.
        """
        self._test_func3([InMemoryDROP("b", "b")], [["b", "c", "d"]])

    def test_func3_multioutput(self):
        """
        Checks that func3 in this module works when wrapped as an application
        with multiple outputs.
        """
        output_drops = [InMemoryDROP(x, x) for x in ("b", "c", "d")]
        self._test_func3(output_drops, ("b", "c", "d"))

    def _test_defaults(self, expected_out, *args, **kwargs):
        def _do_test(func, expected_out, *args, **kwargs):
            n_args = len(args)
            n_kwargs = len(kwargs)
            # List with (drop, value) elements
            arg_inputs = []
            # dict with name: (drop, value) items
            kwarg_inputs = {}
            arg_names = [
                "b",
                "c",
                "x",
                "y",
                "z",
            ]  # neeed to use argument names
            translate = lambda x: base64.b64encode(pickle.dumps(x))
            logger.debug(f"args: {args}")
            for i in range(n_args):
                logger.debug(f"adding arg input: {args[i]}")
                si = arg_names[i]
                arg_inputs.append(InMemoryDROP(si, si, pydata=translate(args[i])))
            i = n_args
            for name, value in kwargs.items():
                si = name  # use keyword name
                kwarg_inputs[name] = (
                    si,
                    InMemoryDROP(si, si, pydata=translate(value)),
                )
                i += 1

            a = InMemoryDROP("a", "a", pydata=translate(1))
            output = InMemoryDROP("o", "o")
            kwargs = {inp.uid: inp.oid for inp in arg_inputs}
            kwargs.update({name: vals[0] for name, vals in kwarg_inputs.items()})
            kwargs["a"] = a.oid
            app = _PyFuncApp(
                "f",
                "f",
                func,
                **kwargs,
            )
            logger.debug(f"adding input: {a}")
            app.addInput(a)
            app.addOutput(output)
            logger.debug(
                f"adding inputs: {arg_inputs + [x[1] for x in kwarg_inputs.values()]}"
            )
            for drop in arg_inputs + [x[1] for x in kwarg_inputs.values()]:
                app.addInput(drop)

            with droputils.DROPWaiterCtx(self, output, timeout=300):
                a.setCompleted()
                for i in arg_inputs + [x[1] for x in kwarg_inputs.values()]:
                    i.setCompleted()

            self.assertEqual(
                expected_out, pickle.loads(droputils.allDropContents(output))
            )  # @UndefinedVariable

        # func_with_defaults returns a - b * c + (y - x) * z
        # defaults are: b=10, c=20, x=30, y=40, z=50
        # The value of "a" is always 1
        _do_test("func_with_defaults", expected_out, *args, **kwargs)
        _do_test(func_with_defaults, expected_out, *args, **kwargs)

    def test_defaults_empty(self):
        # 1 - b * c + (y - x) * z
        # defaults are: b=10, c=20, x=30, y=40, z=50
        self._test_defaults(301)

    def test_defaults_positional_args_only(self):
        # 1 - b * c + (y - x) * z
        # defaults are: b=10, c=20, x=30, y=40, z=50
        # self._test_defaults(501, 0)
        # self._test_defaults(481, 1)
        self._test_defaults(0, 1, 1, 40)
        # self._test_defaults(249, 1, 2, x=35)

    # @unittest.skip
    def test_defaults_kwargs_only(self):
        self._test_defaults(1, z=0, c=0)
        # self._test_defaults(1, z=0, b=0)
        # self._test_defaults(561, b=-1, y=300, z=2)

    # @unittest.skip
    def test_defaults_args_and_kwargs(self):
        self._test_defaults(561, -1, y=300, z=2)
        self._test_defaults(0, 1, 1, x=40)
        self._test_defaults(249, 1, 2, x=35)

    def test_mixed_explicit_and_variable_args(self):
        args = [1, 20, 30]
        kwargs = {"b": 100}
        self.assertEqual(sum_with_args_and_kwarg(1), 1)
        self.assertEqual(sum_with_args_and_kwarg(*args, **kwargs), 151)
        self.assertEqual(sum_with_args_and_kwarg(1, 20, 30, b=100), 151)

    def test_reproducibility(self):
        from dlg.common.reproducibility.constants import ReproducibilityFlags
        from dlg.data.drops.data_base import NullDROP

        a = _PyFuncApp("a", "a", "func3")
        a.run()
        b = NullDROP("b", "b")
        a.reproducibility_level = ReproducibilityFlags.RERUN
        b.reproducibility_level = ReproducibilityFlags.RERUN
        a.setCompleted()
        b.setCompleted()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPEAT
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), {"args": {}})

        a.reproducibility_level = ReproducibilityFlags.REPRODUCE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), {})

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_rerun_data())

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_recompute_data())

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_repeat_data())


class PyFuncAppIntraNMTest(test_dm.NMTestsMixIn, unittest.TestCase):
    def test_input_in_remote_nm(self):
        """
        A test similar in spirit to TestDM.test_runGraphOneDOPerDom, but where
        application B is a PyFuncApp. This makes sure that PyFuncApp work fine
        across Node Managers.

        NM #1      NM #2
        =======    =============
        | A --|----|-> B --> C |
        =======    =============
        """
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        g2 = [
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.pyfunc.PyFuncApp",
                "func_name": __name__ + ".func1",
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "producers": ["B"],
            },
        ]
        rels = [DROPRel("A", DROPLinkType.INPUT, "B")]
        a_data = os.urandom(32)
        c_data = self._test_runGraphInTwoNMs(g1, g2, rels, pickle.dumps(a_data), None)
        self.assertEqual(a_data, pickle.loads(c_data))

    def test_output_in_remote_nm(self):
        """
        Like the above, but with this graph. In this case the output (instead of
        the input) is in a remote Node Manager.

        NM #1            NM #2
        =============    =======
        | A --> B --|----|-> C |
        =============    =======
        """
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "consumers": ["B"],
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.pyfunc.PyFuncApp",
                "func_name": __name__ + ".func1",
            },
        ]
        g2 = [
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        rels = [DROPRel("B", DROPLinkType.PRODUCER, "C")]
        a_data = os.urandom(32)
        c_data = self._test_runGraphInTwoNMs(g1, g2, rels, pickle.dumps(a_data), None)
        self.assertEqual(a_data, pickle.loads(c_data))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A small module that measures the average memory consumption of different
DROP types. It was initially developed to address PRO-234.
"""

import importlib
import sys
from optparse import OptionParser

import psutil


def measure(n, droptype):
    """
    Create `n` DROPs of type `droptype` and measure how much memory does the
    program use at the beginning and the end of the process. It returns a list
    with the total amount of memory, user time and system time used during the
    creation of all the DROP instances
    """
    p = psutil.Process()
    mem1 = p.memory_info()[0]
    uTime1, sTime1 = p.cpu_times()
    drops = []
    for i in xrange(n):
        uid = str(i)
        drops.append(droptype(uid, uid))
    mem2 = p.memory_info()[0]
    uTime2, sTime2 = p.cpu_times()

    return mem2 - mem1, uTime2 - uTime1, sTime2 - sTime1


if __name__ == "__main__":

    parser = OptionParser()
    parser.add_option(
        "--csv",
        action="store_true",
        dest="csv",
        help="Output results in CSV format",
        default=False,
    )
    parser.add_option(
        "-i",
        "--instances",
        action="store",
        type="int",
        dest="instances",
        help="Number of DROP instances to create and measure",
    )
    parser.add_option(
        "-t",
        "--type",
        action="store",
        type="string",
        dest="type",
        help="DROP type to instantiate",
    )
    (options, args) = parser.parse_args(sys.argv)

    if options.type is None:
        parser.error("DROP type to instantiate not specified")
    if options.instances is None:
        parser.error("Number of instances to create not specified")

    n = options.instances
    parts = options.type.split(".")
    modname = ".".join(parts[:-1])
    classname = parts[-1]
    droptype = getattr(importlib.import_module(modname), classname)
    mem, uTime, sTime = measure(n, droptype)
    tTime = uTime + sTime
    memAvg, uTimeAvg, sTimeAvg, tTimeAvg = [
        x / float(n) for x in (mem, uTime, sTime, tTime)
    ]

    if options.csv:
        print(
            "%s,%d,%d,%.2f,%.2f,%.2f,%.2f,%.2f,%.2f,%.2f"
            % (
                options.type,
                n,
                mem,
                uTime * 1e3,
                sTime * 1e3,
                tTime * 1e3,
                memAvg,
                uTimeAvg * 1e6,
                sTimeAvg * 1e6,
                tTimeAvg * 1e6,
            )
        )
    else:
        print(
            "%d bytes used by %d %ss (%.2f bytes per DROP)"
            % (mem, n, droptype.__name__, memAvg)
        )
        print("Total time:  %.2f msec (%.2f msec per DROP)" % (tTime, tTimeAvg))
        print("User time:   %.2f msec (%.2f msec per DROP)" % (uTime, uTimeAvg))
        print("System time: %.2f msec (%.2f msec per DROP)" % (sTime, sTimeAvg))



#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2020
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

# Try to compile the library, if possible. If it's there already we're cool
def build_shared_library(libname, libpath):
    import os
    import distutils.ccompiler
    from dlg.runtime import get_include_dir

    prev_path = os.getcwd()
    os.chdir(os.path.dirname(__file__))

    try:

        # No need to rebuild
        srcname = libname + ".c"
        if (
            os.path.isfile(libpath)
            and os.stat(srcname).st_ctime <= os.stat(libpath).st_ctime
        ):
            return True

        comp = distutils.ccompiler.new_compiler()
        distutils.sysconfig.customize_compiler(comp)
        comp.add_include_dir(distutils.sysconfig.get_python_inc())

        comp.add_include_dir(get_include_dir())
        objs = comp.compile([srcname])
        comp.link_shared_lib(objs, output_libname=libname)

        return True
    except:
        return False
    finally:
        os.chdir(prev_path)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the REST layer that exposes the methods of the different
Data Managers (DROPManager and DataIslandManager) to the outside world.
"""

import cgi
import functools
import io
import json
import logging
import os
import re
import tarfile
import threading

import bottle
import pkg_resources

from bottle import static_file

from . import constants
from .client import NodeManagerClient, DataIslandManagerClient
from .. import utils
from ..exceptions import (
    InvalidGraphException,
    InvalidSessionState,
    DaliugeException,
    NoSessionException,
    SessionAlreadyExistsException,
    InvalidDropException,
    InvalidRelationshipException,
    SubManagerException,
)
from ..restserver import RestServer
from ..restutils import RestClient, RestClientException
from .session import generateLogFileName
from dlg.common.deployment_methods import DeploymentMethods

logger = logging.getLogger(__name__)


def file_as_string(fname, enc="utf8"):
    b = pkg_resources.resource_string(__name__, fname)  # @UndefinedVariable
    return utils.b2s(b, enc)


def daliuge_aware(func):
    @functools.wraps(func)
    def fwrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)

            if isinstance(res, (bytes, bottle.HTTPResponse)):
                return res

            if res is not None:
                bottle.response.content_type = "application/json"
                # set CORS headers
                origin = bottle.request.headers.raw("Origin")
                if origin is None:
                    origin = "http://localhost:8084"
                elif not re.match(
                    r"http://((localhost)|(127.0.0.1)):80[0-9][0-9]", origin
                ):
                    origin = "http://localhost:8084"
                bottle.response.headers["Access-Control-Allow-Origin"] = origin
                bottle.response.headers[
                    "Access-Control-Allow-Credentials"
                ] = "true"
                bottle.response.headers[
                    "Access-Control-Allow-Methods"
                ] = "GET, POST, PUT, OPTIONS"
                bottle.response.headers[
                    "Access-Control-Allow-Headers"
                ] = "Origin, Accept, Content-Type, Content-Encoding, X-Requested-With, X-CSRF-Token"
            jres = (
                json.dumps(res) if res else json.dumps({"Status": "Success"})
            )
            logger.debug(
                "Bottle sending back result: %s", jres[: min(len(jres), 80)]
            )
            return json.dumps(res)
        except Exception as e:
            logger.exception("Error while fulfilling request")

            status, eargs = 500, ()
            if isinstance(e, NotImplementedError):
                status, eargs = 501, e.args
            elif isinstance(e, NoSessionException):
                status, eargs = 404, (e._session_id,)
            elif isinstance(e, SessionAlreadyExistsException):
                status, eargs = 409, (e._session_id,)
            elif isinstance(e, InvalidDropException):
                status, eargs = 409, ((e.oid, e.uid), e.reason)
            elif isinstance(e, InvalidRelationshipException):
                status, eargs = 409, (e.rel, e.reason)
            elif isinstance(e, InvalidGraphException):
                status, eargs = 400, e.args
            elif isinstance(e, InvalidSessionState):
                status, eargs = 400, e.args
            elif isinstance(e, RestClientException):
                status, eargs = 556, e.args
            elif isinstance(e, SubManagerException):
                status = 555
                eargs = {}
                # args[1] is a dictionary of host:exception
                for host, subex in e.args[1].items():
                    eargs[host] = {
                        "type": subex.__class__.__name__,
                        "args": subex.args,
                    }
            elif isinstance(e, DaliugeException):
                status, eargs = 555, e.args
            else:
                raise

            error = {"type": e.__class__.__name__, "args": eargs}
            bottle.response.status = status
            return json.dumps(error)

    return fwrapper


class ManagerRestServer(RestServer):
    """
    An object that wraps a DataManager and exposes its methods via a REST
    interface. The server is started via the `start` method in a separate thread
    and runs until the process is shut down.

    This REST server currently also serves HTML pages in some of its methods
    (i.e. those not under /api).
    """

    def __init__(self, dm, maxreqsize=10):
        super(ManagerRestServer, self).__init__()

        # Increase maximum file sizes
        bottle.BaseRequest.MEMFILE_MAX = maxreqsize * 1024 * 1024

        self.dm = dm

        # Mappings
        app = self.app
        app.get("/api/submission_method", callback=self.submit_methods)
        app.post("/api/stop", callback=self.stop_manager)
        app.post("/api/sessions", callback=self.createSession)
        app.get("/api/sessions", callback=self.getSessions)
        app.get(
            "/api/sessions/<sessionId>", callback=self.getSessionInformation
        )
        app.delete("/api/sessions/<sessionId>", callback=self.destroySession)
        app.get("/api/sessions/<sessionId>/logs", callback=self.getLogFile)
        app.get(
            "/api/sessions/<sessionId>/status", callback=self.getSessionStatus
        )
        app.post(
            "/api/sessions/<sessionId>/deploy", callback=self.deploySession
        )
        app.post(
            "/api/sessions/<sessionId>/cancel", callback=self.cancelSession
        )
        app.get("/api/sessions/<sessionId>/graph", callback=self.getGraph)
        app.get(
            "/api/sessions/<sessionId>/graph/size", callback=self.getGraphSize
        )
        app.get(
            "/api/sessions/<sessionId>/graph/status",
            callback=self.getGraphStatus,
        )
        app.post(
            "/api/sessions/<sessionId>/graph/append",
            callback=self.addGraphParts,
        )
        app.get(
            "/api/sessions/<sessionId>/repro/data",
            callback=self.getSessionReproData,
        )
        app.get(
            "/api/sessions/<sessionId>/repro/status",
            callback=self.getSessionReproStatus,
        )

        app.route(
            "/api/sessions", method="OPTIONS", callback=self.acceptPreflight
        )
        app.route(
            "/api/sessions/<sessionId>/graph/append",
            method="OPTIONS",
            callback=self.acceptPreflight2,
        )

        # The non-REST mappings that serve HTML-related content
        app.route("/static/<filepath:path>", callback=self.server_static)
        app.get("/session", callback=self.visualizeSession)

        # sub-class specifics
        self.initializeSpecifics(app)

    def initializeSpecifics(self, app):
        """
        Methods through which subclasses can initialize other mappings on top of
        the default ones and perform other DataManager-specific actions.
        The default implementation does nothing.
        """

    @daliuge_aware
    def submit_methods(self):
        return {
            "methods": [DeploymentMethods.BROWSER, DeploymentMethods.SERVER]
        }

    def _stop_manager(self):
        self.dm.shutdown()
        self.stop()
        logger.info(
            "Thanks for using our %s, come back again :-)",
            self.dm.__class__.__name__,
        )

    @daliuge_aware
    def stop_manager(self):
        threading.Thread(target=self._stop_manager).start()

    @daliuge_aware
    def createSession(self):
        newSession = bottle.request.json
        sessionId = newSession["sessionId"]
        self.dm.createSession(sessionId)
        return {"sessionId": sessionId}

    @daliuge_aware
    def acceptPreflight(self):
        return {}

    @daliuge_aware
    def acceptPreflight2(self, sessionId):
        return {}

    def sessions(self):
        sessions = []
        for sessionId in self.dm.getSessionIds():
            sessions.append(
                {
                    "sessionId": sessionId,
                    "status": self.dm.getSessionStatus(sessionId),
                    "size": self.dm.getGraphSize(sessionId),
                }
            )
        return sessions

    @daliuge_aware
    def getSessions(self):
        return self.sessions()

    @daliuge_aware
    def getSessionInformation(self, sessionId):
        status = self.dm.getSessionStatus(sessionId)
        try:
            graphDict = self.dm.getGraph(sessionId)
        except:  # Pristine state sessions don't have a graph, yet.
            graphDict = {}
            status = 0
        return {"status": status, "graph": graphDict}

    @daliuge_aware
    def getSessionReproStatus(self, sessionId):
        return self.dm.getSessionReproStatus(sessionId)

    @daliuge_aware
    def getSessionsReproStatus(self):
        sessions = []
        for sessionId in self.dm.getSessionIds():
            sessions.append(
                {
                    "sessionId": sessionId,
                    "status": self.dm.getSessionStatus(sessionId),
                    "size": self.dm.getGraphSize(sessionId),
                    "repro": self.dm.getSessionReproStatus(sessionId),
                }
            )
        return sessions

    @daliuge_aware
    def getSessionReproData(self, sessionId):
        #  For now, we only have information on a per-graph basis.
        graphDict = self.dm.getGraph(sessionId)
        reprodata = self.dm.getGraphReproData(sessionId)
        return {"graph": graphDict, "reprodata": reprodata}

    @daliuge_aware
    def destroySession(self, sessionId):
        self.dm.destroySession(sessionId)

    @daliuge_aware
    def getSessionStatus(self, sessionId):
        return self.dm.getSessionStatus(sessionId)

    @daliuge_aware
    def deploySession(self, sessionId):
        completedDrops = []
        if "completed" in bottle.request.forms:
            completedDrops = bottle.request.forms["completed"].split(",")
        return self.dm.deploySession(sessionId, completedDrops=completedDrops)
        # return {"Status": "Success"}

    @daliuge_aware
    def cancelSession(self, sessionId):
        self.dm.cancelSession(sessionId)

    @daliuge_aware
    def getGraph(self, sessionId):
        return self.dm.getGraph(sessionId)

    @daliuge_aware
    def getGraphSize(self, sessionId):
        return self.dm.getGraphSize(sessionId)

    @daliuge_aware
    def getGraphStatus(self, sessionId):
        return self.dm.getGraphStatus(sessionId)

    # TODO: addGraphParts v/s addGraphSpec
    @daliuge_aware
    def addGraphParts(self, sessionId):
        # WARNING: TODO: Somehow, the content_type can be overwritten to 'text/plain'
        logger.debug("Graph content type: %s", bottle.request.content_type)
        if (
            "application/json" not in bottle.request.content_type
            and "text/plain" not in bottle.request.content_type
        ):
            bottle.response.status = 415
            return

        # We also accept gzipped content
        hdrs = bottle.request.headers
        logger.debug("Graph hdr: %s", {k: v for k, v in hdrs.items()})
        if hdrs.get("Content-Encoding", None) == "gzip":
            json_content = utils.ZlibUncompressedStream(bottle.request.body)
        else:
            json_content = bottle.request.body

        graph_parts = bottle.json_loads(json_content.read())

        return self.dm.addGraphSpec(sessionId, graph_parts)
        # return {"graph_parts": graph_parts}

    # ===========================================================================
    # non-REST methods
    # ===========================================================================
    def server_static(self, filepath):
        staticRoot = pkg_resources.resource_filename(
            __name__, "/web/static"
        )  # @UndefinedVariable
        return bottle.static_file(filepath, root=staticRoot)

    def visualizeSession(self):
        params = bottle.request.params
        sessionId = params["sessionId"] if "sessionId" in params else ""
        selectedNode = params["node"] if "node" in params else ""
        viewMode = params["view"] if "view" in params else ""
        tpl = file_as_string("web/session.html")
        urlparts = bottle.request.urlparts
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            sessionId=sessionId,
            selectedNode=selectedNode,
            viewMode=viewMode,
            serverUrl=serverUrl,
            dmType=self.dm.__class__.__name__,
        )


class NMRestServer(ManagerRestServer):
    """
    A REST server for NodeManagers. It includes mappings for NM-specific
    methods and the mapping for the main visualization HTML pages.
    """

    def initializeSpecifics(self, app):
        app.get("/api", callback=self.getNMStatus)
        app.post(
            "/api/sessions/<sessionId>/graph/link",
            callback=self.linkGraphParts,
        )
        app.post(
            "/api/sessions/<sessionId>/subscriptions",
            callback=self.add_node_subscriptions,
        )
        app.post(
            "/api/sessions/<sessionId>/trigger", callback=self.trigger_drops
        )
        # The non-REST mappings that serve HTML-related content
        app.get("/", callback=self.visualizeDM)
        app.get("/api/shutdown", callback=self.shutdown_node_manager)

    @daliuge_aware
    def shutdown_node_manager(self):
        logger.debug("Shutting down node manager")
        self.dm.shutdown()

    @daliuge_aware
    def getNMStatus(self):
        # we currently return the sessionIds, more things might be added in the
        # future
        logger.debug("NM REST call: status")
        return {"sessions": self.sessions()}

    @daliuge_aware
    def getLogFile(self, sessionId):
        logger.debug("NM REST call: logfile")
        logdir = self.dm.getLogDir()
        logfile = generateLogFileName(logdir, sessionId)
        if not os.path.isfile(logfile):
            raise NoSessionException(sessionId, "Log file not found.")
        return static_file(
            os.path.basename(logfile),
            root=logdir,
            download=os.path.basename(logfile),
        )

    @daliuge_aware
    def linkGraphParts(self, sessionId):
        logger.debug("NM REST call: graph/link")
        params = bottle.request.params
        lhOID = params["lhOID"]
        rhOID = params["rhOID"]
        linkType = int(params["linkType"])
        self.dm.linkGraphParts(sessionId, lhOID, rhOID, linkType)

    @daliuge_aware
    def add_node_subscriptions(self, sessionId):
        logger.debug("NM REST call: add_subscriptions %s", bottle.request.json)
        if bottle.request.content_type != "application/json":
            bottle.response.status = 415
            return
        self.dm.add_node_subscriptions(sessionId, bottle.request.json)

    @daliuge_aware
    def trigger_drops(self, sessionId):
        if bottle.request.content_type != "application/json":
            bottle.response.status = 415
            return
        self.dm.trigger_drops(sessionId, bottle.request.json)

    # ===========================================================================
    # non-REST methods
    # ===========================================================================
    def visualizeDM(self):
        tpl = file_as_string("web/dm.html")
        urlparts = bottle.request.urlparts
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            serverUrl=serverUrl,
            dmType=self.dm.__class__.__name__,
            reset="false",
        )


class CompositeManagerRestServer(ManagerRestServer):
    """
    A REST server for DataIslandManagers. It includes mappings for DIM-specific
    methods.
    """

    def initializeSpecifics(self, app):
        app.get("/api", callback=self.getCMStatus)
        app.get("/api/nodes", callback=self.getCMNodes)
        app.post("/api/node/<node>", callback=self.addCMNode)
        app.delete("/api/node/<node>", callback=self.removeCMNode)

        # Query forwarding to sub-nodes
        app.get("/api/node/<node>/sessions", callback=self.getNodeSessions)
        app.get(
            "/api/node/<node>/sessions/<sessionId>",
            callback=self.getNodeSessionInformation,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/status",
            callback=self.getNodeSessionStatus,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/graph",
            callback=self.getNodeGraph,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/graph/status",
            callback=self.getNodeGraphStatus,
        )

        # The non-REST mappings that serve HTML-related content
        app.get("/", callback=self.visualizeDIM)

    @daliuge_aware
    def getCMStatus(self):
        return {
            "hosts": self.dm.dmHosts,
            "sessionIds": self.dm.getSessionIds(),
        }

    @daliuge_aware
    def getCMNodes(self):
        return self.dm.nodes

    def getAllCMNodes(self):
        return self.dm.nodes

    @daliuge_aware
    def addCMNode(self, node):
        logger.debug("Adding node %s", node)
        self.dm.add_node(node)

    @daliuge_aware
    def removeCMNode(self, node):
        logger.debug("Removing node %s", node)
        self.dm.remove_node(node)

    @daliuge_aware
    def getNodeSessions(self, node):
        if node not in self.dm.nodes:
            raise Exception(f"{node} not in current list of nodes")
        with NodeManagerClient(host=node) as dm:
            return dm.sessions()

    def _tarfile_write(self, tar, headers, stream):
        file_header = headers.getheader("Content-Disposition")
        length = headers.getheader("Content-Length")
        _, params = cgi.parse_header(file_header)
        filename = params["filename"]
        info = tarfile.TarInfo(filename)
        info.size = int(length)

        content = []
        while True:
            buffer = stream.read()
            if not buffer:
                break
            content.append(buffer)

        tar.addfile(info, io.BytesIO(initial_bytes="".join(content).encode()))

    @daliuge_aware
    def getLogFile(self, sessionId):
        fh = io.BytesIO()
        with tarfile.open(fileobj=fh, mode="w:gz") as tar:
            for node in self.getAllCMNodes():
                with NodeManagerClient(host=node) as dm:
                    try:
                        stream, resp = dm.get_log_file(sessionId)
                        self._tarfile_write(tar, resp, stream)
                    except NoSessionException:
                        pass

        data = fh.getvalue()
        size = len(data)
        bottle.response.set_header("Content-type", "application/x-tar")
        bottle.response["Content-Disposition"] = (
            f"attachment; " f"filename=dlg_{sessionId}.tar"
        )
        bottle.response["Content-Length"] = size
        return data

    @daliuge_aware
    def getNodeSessionInformation(self, node, sessionId):
        if node not in self.dm.nodes:
            raise Exception(f"{node} not in current list of nodes")
        with NodeManagerClient(host=node) as dm:
            return dm.session(sessionId)

    @daliuge_aware
    def getNodeSessionStatus(self, node, sessionId):
        if node not in self.dm.nodes:
            raise Exception(f"{node} not in current list of nodes")
        with NodeManagerClient(host=node) as dm:
            return dm.session_status(sessionId)

    @daliuge_aware
    def getNodeGraph(self, node, sessionId):
        if node not in self.dm.nodes:
            raise Exception(f"{node} not in current list of nodes")
        with NodeManagerClient(host=node) as dm:
            return dm.graph(sessionId)

    @daliuge_aware
    def getNodeGraphStatus(self, node, sessionId):
        if node not in self.dm.nodes:
            raise Exception(f"{node} not in current list of nodes")
        with NodeManagerClient(host=node) as dm:
            return dm.graph_status(sessionId)

    # ===========================================================================
    # non-REST methods
    # ===========================================================================
    def visualizeDIM(self):
        tpl = file_as_string("web/dim.html")
        urlparts = bottle.request.urlparts
        selectedNode = (
            bottle.request.params["node"]
            if "node" in bottle.request.params
            else ""
        )
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            dmType=self.dm.__class__.__name__,
            dmPort=self.dm.dmPort,
            serverUrl=serverUrl,
            dmHosts=json.dumps(self.dm.dmHosts),
            nodes=json.dumps(self.dm.nodes),
            selectedNode=selectedNode,
        )


class MasterManagerRestServer(CompositeManagerRestServer):
    def initializeSpecifics(self, app):
        CompositeManagerRestServer.initializeSpecifics(self, app)
        # DIM manamagement
        app.post("/api/island/<dim>", callback=self.addDIM)
        app.delete("/api/island/<dim>", callback=self.removeDIM)
        # Query forwarding to daemons
        app.post("/api/managers/<host>/island", callback=self.createDataIsland)
        app.post("/api/managers/<host>/node/start", callback=self.startNM)
        app.post("/api/managers/<host>/node/stop", callback=self.stopNM)
        # manage node manager assignment
        app.post("/api/managers/<host>/node/<node>", callback=self.addNM)
        app.delete("/api/managers/<host>/node/<node>", callback=self.removeNM)
        # Querying about managers
        app.get("/api/islands", callback=self.getDIMs)
        app.get("/api/nodes", callback=self.getNMs)
        app.get("/api/managers/<host>/node", callback=self.getNMInfo)
        app.get("/api/managers/<host>/island", callback=self.getDIMInfo)
        app.get("/api/managers/<host>/master", callback=self.getMMInfo)

    @daliuge_aware
    def createDataIsland(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            c._post_json("/managers/island/start", bottle.request.body.read())
        self.dm.addDmHost(host)
        return {"islands": self.dm.dmHosts}

    @daliuge_aware
    def getDIMs(self):
        return {"islands": self.dm.dmHosts}

    @daliuge_aware
    def addDIM(self, dim):
        logger.debug("Adding DIM %s", dim)
        self.dm.addDmHost(dim)

    @daliuge_aware
    def removeDIM(self, dim):
        logger.debug("Removing dim %s", dim)
        self.dm.removeDmHost(dim)

    @daliuge_aware
    def getNMs(self):
        return {"nodes": self.dm.nodes}

    @daliuge_aware
    def startNM(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending NM start request to %s:%s", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c._POST("/managers/node/start").read())

    @daliuge_aware
    def stopNM(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending NM stop request to %s:%s", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c._POST("/managers/node/stop").read())

    @daliuge_aware
    def addNM(self, host, node):
        port = constants.ISLAND_DEFAULT_REST_PORT
        logger.debug("Adding NM %s to DIM %s", node, host)
        with RestClient(
            host=host, port=port, timeout=10, url_prefix="/api"
        ) as c:
            return json.loads(
                c._POST(
                    f"/node/{node}",
                ).read()
            )

    @daliuge_aware
    def removeNM(self, host, node):
        port = constants.ISLAND_DEFAULT_REST_PORT
        logger.debug("Removing NM %s from DIM %s", node, host)
        with RestClient(
            host=host, port=port, timeout=10, url_prefix="/api"
        ) as c:
            return json.loads(c._DELETE(f"/node/{node}").read())

    @daliuge_aware
    def getNMInfo(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending request %s:%s/managers/node", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c._GET("/managers/node").read())

    @daliuge_aware
    def getDIMInfo(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            return json.loads(c._GET("/managers/island").read())

    @daliuge_aware
    def getMMInfo(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            return json.loads(c._GET("/managers/master").read())

    def getAllCMNodes(self):
        nodes = []
        for node in self.dm.dmHosts:
            with DataIslandManagerClient(host=node) as dm:
                nodes += dm.nodes()
        return nodes



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

from asyncio.log import logger
import os
import shutil
import tempfile
import unittest

import configobj
import docker

from dlg import droputils, utils, prepareUser
from dlg.apps.dockerapp import DockerApp
from dlg.data.drops.ngas import NgasDROP
from dlg.data.drops.file import FileDROP
from dlg.droputils import DROPWaiterCtx

docker_unavailable = True
try:
    docker.from_env(version="auto").ping()
    docker_unavailable = False
except:
    pass


@unittest.skipIf(docker_unavailable, "Docker daemon not available")
class DockerTests(unittest.TestCase):
    _temp = None
    _docker_available = False

    @classmethod
    def setUpClass(cls):
        config_file_name = os.path.join(utils.getDlgDir(), "dlg.settings")
        if os.path.exists(config_file_name):
            config = configobj.ConfigObj(config_file_name)
            cls._temp = config.get("OS_X_TEMP")

            if not os.path.exists(DockerTests._temp):
                os.makedirs(DockerTests._temp)

        if cls._temp is None:
            cls._temp = "/tmp/daliuge_tfiles"

        os.environ["DLG_ROOT"] = cls._temp
        logger.info(f"Preparing pwd and group files in {utils.getDlgDir()}")
        _dum = prepareUser.prepareUser(DLG_ROOT=utils.getDlgDir())

    @classmethod
    def tearDownClass(cls):
        logger.debug(f"Removing temp directory {cls._temp}")
        shutil.rmtree(cls._temp, True)

    def test_simpleCopy(self):
        """
        Simple test for a dockerized application. It copies the contents of one
        file into another via the command-line cp utility. It then checks that
        the contents of the target DROP are correct, and that the target file is
        actually owned by our process.
        """

        a = FileDROP("a", "a")
        b = DockerApp("b", "b", image="ubuntu:14.04", command="cp %i0 %o0")
        c = FileDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        # Random data so we always check different contents
        data = os.urandom(10)
        with DROPWaiterCtx(self, c, 100):
            a.write(data)
            a.setCompleted()

        self.assertEqual(data, droputils.allDropContents(c))

        # We own the file, not root
        uid = os.getuid()
        self.assertEqual(uid, os.stat(c.path).st_uid)

    def test_clientServer(self):
        """
        A client-server duo. The server outputs the data it receives to its
        output DROP, which in turn is the data held in its input DROP. The graph
        looks like this:

        A --|--> B(client) --|--> D
            |--> C(server) --|

        C is a server application which B connects to. Therefore C must be
        started before B, so B knows C's IP address and connects successfully.
        Although the real writing is done by C, B in this example is also
        treated as a publisher of D. This way D waits for both applications to
        finish before proceeding.
        """

        a = FileDROP("a", "a")
        b = DockerApp(
            "b",
            "b",
            image="ubuntu:14.04",
            command="cat %i0 > /dev/tcp/%containerIp[c]%/8000",
        )
        c = DockerApp("c", "c", image="ubuntu:14.04", command="nc -l 8000 > %o0")
        d = FileDROP("d", "d")

        b.addInput(a)
        b.addOutput(d)
        c.addInput(a)
        c.addOutput(d)

        # Let 'b' handle its interest in c
        b.handleInterest(c)

        data = os.urandom(10)
        with DROPWaiterCtx(self, d, 10):
            a.write(data)
            a.setCompleted()

        self.assertEqual(data, droputils.allDropContents(d))

    def test_quotedCommands(self):
        """
        A test to check that commands using quotes are correctly executed, which
        means that their quotes were correctly escaped when the final docker
        command was executed
        """

        def assertMsgIsCorrect(msg, command):
            a = DockerApp("a", "a", image="ubuntu:14.04", command=command)
            b = FileDROP("b", "b")
            a.addOutput(b)
            with DROPWaiterCtx(self, b, 100):
                a.execute()
            self.assertEqual(msg.encode("utf8"), droputils.allDropContents(b))

        msg = "This is a message with a single quote: '"
        assertMsgIsCorrect(msg, 'echo -n "{0}" > %o0'.format(msg))
        msg = 'This is a message with a double quotes: "'
        assertMsgIsCorrect(msg, "echo -n '{0}' > %o0".format(msg))

    @unittest.skip
    def test_dataURLReference(self):
        """
        A test to check that DROPs other than FileDROPs and DirectoryContainers
        can pass their dataURLs into docker containers
        """
        self._ngas_and_fs_io("echo -n '%iDataURL0' > %o0")

    @unittest.skip
    def test_refer_to_io_by_uid(self):
        """
        A test to check that input and output Drops can be referred to by their
        UIDs (in addition to their position in the list of inputs or outputs)
        in the command-line.
        """
        self._ngas_and_fs_io("echo -n '%iDataURL[HelloWorld_out.txt]' > %o[c]")

    def _ngas_and_fs_io(self, command):
        a = NgasDROP(
            "HelloWorld_out.txt", "HelloWorld_out.txt"
        )  # not a filesystem-related DROP, we can reference its URL in the command-line
        a.ngasSrv = "ngas.icrar.org"
        b = DockerApp("b", "b", image="ubuntu:14.04", command=command)
        c = FileDROP("c", "c")
        b.addInput(a)
        b.addOutput(c)
        with DROPWaiterCtx(self, c, 100):
            a.setCompleted()
        self.assertEqual(a.dataURL.encode("utf8"), droputils.allDropContents(c))

    def test_additional_bindings(self):
        # Some additional stuff to bind into docker
        tempDir = tempfile.mkdtemp()
        tempFile = tempfile.mktemp()
        with open(tempFile, "w") as f:
            f.write("data")

        # One binding specifies the target path in the container, the other doesn't
        # so it defaults to the same path
        a = DockerApp(
            "a",
            "a",
            image="ubuntu:14.04",
            command="cp /opt/file %s" % (tempDir,),
            additionalBindings=[tempDir, "%s:/opt/file" % (tempFile,)],
        )
        a.execute()

        # We copied the file into the directory, but since in the container the
        # file was called "file" we'll see it with that name in tempDir
        self.assertEqual(1, len(os.listdir(tempDir)))
        with open(os.path.join(tempDir, "file")) as f:
            data = f.read()
        self.assertEqual("data", data)

        # Cleanup
        os.unlink(tempFile)
        shutil.rmtree(tempDir)

    def _test_working_dir(self, ensureUserAndSwitch):
        # the sleep is required to make sure that the docker container exists long enough for
        # DALiuGE to get the required information (2 lines of Python code after starting the container!)
        a = DockerApp(
            "a",
            "a",
            workingDir="/mydir",
            image="ubuntu:14.04",
            command="pwd > %o0 && sleep 0.05",
            ensureUserAndSwitch=ensureUserAndSwitch,
        )
        b = FileDROP("b", "b")
        a.addOutput(b)
        with DROPWaiterCtx(self, b, 10):
            a.execute()
        self.assertEqual(b"/mydir", droputils.allDropContents(b).strip())

    def test_working_dir(self):
        self._test_working_dir(True)
        self._test_working_dir(False)
